#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import ast
import builtins
import copy
import fnmatch
import os
import re
import shutil
import sys
import textwrap
from io import StringIO, TextIOBase
from tokenize import generate_tokens
import typing
from typing import Dict, List, Optional, Tuple, Any, Set
from uuid import UUID, uuid4

from .eval import on_demand_options
from .syntax import (INDENTED, SOS_CELL, SOS_DIRECTIVE, SOS_DIRECTIVES,
                     SOS_FORMAT_LINE, SOS_FORMAT_VERSION, SOS_INDENTED_ACTION,
                     SOS_MAGIC, SOS_SECTION_HEADER, SOS_SECTION_NAME,
                     SOS_SECTION_OPTION, SOS_SUBWORKFLOW, SOS_ACTION_OPTIONS)
from .targets import file_target, path, paths, sos_targets, textMD5
from .utils import Error, env, locate_script, as_fstring, valid_expr_till, format_par, separate_options

__all__ = ['SoS_Script']

# these are needed by eval as recognizable types
assert file_target
assert path
assert paths
assert sos_targets


class ParsingError(Error):
    '''Raised when a configuration file does not follow legal syntax.'''

    def __init__(self, filename: str) -> None:
        Error.__init__(
            self,
            f'File contains parsing errors: {filename if filename != "<string>" else ""}'
        )
        self.filename = filename
        self.errors: List[Tuple[int, str]] = []
        self.args = (filename,)

    def append(self, lineno: int, line: str, msg: str) -> None:
        if (lineno, line) in self.errors:
            return
        self.errors.append((lineno, line))
        self.message += f'\n\t[line {lineno:2d}]: {line}\n{msg}'


_action_list = None


def is_type_hint(stmt: str) -> bool:
    '''Try to differentiate

    var: type = value

    with

    action: input = whatever

    '''
    if stmt.count('=') > 1:
        return False

    if ':' not in stmt:
        return False

    #
    # action:
    if not stmt.split(':')[1].strip():
        return False

    #
    # action: int
    #
    # or
    #
    # input: variable
    #
    if '=' not in stmt:
        action, par = [x.strip() for x in stmt.split(':', 1)]
    else:
        # one parameter?
        #
        # action: input={'a': b}
        #
        action, par = [x.strip() for x in stmt.split('=', 1)[0].split(':', 1)]

    if action in SOS_DIRECTIVES:
        return False
    if par in SOS_ACTION_OPTIONS:
        return False

    # if par is something like List[Any], or 'classname'
    if not par.isidentifier():
        return True

    # if action is a builtin function, such as sort, it cannot be
    # a variable assignment.
    if action in dir(builtins):
        return False

    # if action is registered
    global _action_list
    if _action_list is None:
        import pkg_resources
        _action_list = [
            x.name for x in pkg_resources.iter_entry_points(group='sos_actions')
        ]
    if action in _action_list:
        return False

    # if par is something like List, Tuple, str
    if par in dir(typing) or par in dir(builtins):
        return True

    # if not quite sure???
    env.logger.debug(
        f"Failed to tell if '{stmt}' is an assignment with type hint or function in script format. Assuming type hint."
    )
    # regular function written in this format?
    return True


def extract_option_from_arg_list(options: str, optname: str,
                                 default_value: None) -> Tuple[Any, str]:
    if not options:
        return default_value, options
    try:
        args = list(ast.iter_fields(ast.parse(f"f({options})",
                                              mode='eval')))[0][1].keywords
        for field in args:
            if field.arg == optname:
                try:
                    value = eval(
                        compile(
                            ast.Expression(body=field.value),
                            filename="<ast>",
                            mode="eval"))
                    new_options = ','.join([
                        x for x in options.split(',')
                        if not x.strip().startswith(optname)
                    ])
                    return value, new_options.strip()
                except Exception:
                    raise ValueError(
                        f"A constant value is expected for option {optname}: {options} provided."
                    )
        return default_value, options
    except SyntaxError:
        raise ValueError(
            f"Expect a list of keyword arguments: {options} provided")


def replace_sigil(text: str, sigil: str) -> str:
    if sigil == '{ }':
        return text
    if sigil is not None and (sigil.count(' ') != 1 or sigil[0] in (' ', "'") or
                              sigil[-1] in (' ', "'") or
                              sigil.split(' ')[0] == sigil.split(' ')[1]):
        raise ValueError(f'Incorrect sigil "{sigil}"')
    # then we need to replace left sigil as { and right sigil asn }
    l, r = sigil.split(' ')
    # now that we have the correct sigil
    # first, we need to replace all { as {{ and } as }}
    lp = re.compile(re.escape(l))
    rp = re.compile(re.escape(r))
    final_text = ''
    while True:
        pieces = lp.split(text, 1)
        if len(pieces) == 1:
            # cannot split
            final_text += text.replace('{', '{{').replace('}', '}}')
            break
        else:
            rhs = pieces[1]
            # now if there is {, we need to find matching number of }
            # basically we need to find syntaxly valid expression before
            # ending bracket... or !
            rhs_pieces = rp.split(rhs)
            if len(rhs_pieces) == 1:
                raise ValueError(
                    f'Invalid f-string {text}: missing right sigil at {rhs[:20]}'
                )

            # say we have sigil = [ ]
            #
            # rhs = 'whatever ]' :r] something else [ ]
            #
            # we need to include ] in expression
            #
            for n in range(1, len(rhs_pieces) + 1):
                if valid_expr_till(r.join(rhs_pieces[:n])) > 0:
                    final_text += pieces[0].replace('{', '{{').replace('}', '}}') + \
                        '{' + r.join(rhs_pieces[:n]) + '}'
                    text = r.join(rhs_pieces[n:])
                    break
                # the last one, still not valid
                if n == len(rhs_pieces):
                    raise ValueError(
                        f'Invalid f-string {text}: invalid expression')
    # finally, replace LSIGIL etc
    return final_text


def get_names_of_kwargs(param_list):
    tree = ast.parse(f'__null_func__({param_list})')
    return [x.arg for x in tree.body[0].value.keywords]


class SoS_Step:
    '''Parser of a SoS step. This class accepts strings sent by the parser, determine
    their types and add them to appropriate sections (directive, statement,
    scripts etc) '''

    def __init__(self,
                 context: Optional['SoS_ScriptContent'] = None,
                 names=None,
                 options=None,
                 is_global: bool = False,
                 comment: str = '') -> None:
        '''A sos step '''
        self.context = context
        # A step will not have a name and index until it is copied to separate workflows
        self.name = None
        self.index = None
        self.alias = None
        # it initially hold multiple names with/without wildcard characters
        self.names = [] if names is None else names
        # everything before step process
        self.statements: List = []
        self.global_parameters: Dict = {}
        self.parameters: Dict = {}
        self.substep_parameters: Set = set()
        # step processes
        self.global_stmts = ''
        self.global_def = ''
        self.global_vars = {}
        self.task = ''
        self.task_params = ''
        self.last_step = None
        self.comment = comment
        # is it global section? This is a temporary indicator because the global section
        # will be inserted to each step of the workflow.
        self.is_global = is_global
        # indicate the type of input of the last line
        self.values: List = []
        self.lineno = 0
        #
        self.runtime_options: Dict = {}
        self.options = on_demand_options(options)
        #
        # string mode to collect all strings as part of an action
        self._action = None
        self._action_options = ''
        self._script = ''

    def has_external_task(self) -> bool:
        return self.task != ''

    # def has_nested_workflow(self) -> bool:
    #     return any('sos_run' in x[1] for x in self.statements)

    def step_name(self, alias: bool = False) -> str:
        # if the step is not part of any workflow and is not given a name
        if self.name is None:
            names = []
            for n, i, a in self.names:
                if alias and a:
                    names.append(a)
                elif n and i is not None:
                    names.append(f'{n}_{i}')
                else:
                    names.append(n if n is not None else str(i))
            return ', '.join(names)
        else:
            if alias and self.alias:
                return self.alias
            elif self.name and self.index is not None:
                return f'{self.name}_{self.index}'
            else:
                return self.name if self.name else str(self.index)

    def match(self, step_name: str) -> bool:
        # if this step provides name...
        for name, index, _ in self.names:
            if step_name == name or step_name == f'{name}_{0 if index is None else int(index)}':
                return True
        return False

    def indented_script(self) -> bool:
        ''' check self._script and see if it is indented '''
        # get all leading space, tab and newline
        leading = INDENTED.match(self._script)
        return 0 if leading is None else len(leading.group(2))

    def category(self) -> Optional[str]:
        '''Determine the category of existing statement'''
        if self.statements:
            if self.statements[-1][0] == ':':
                # a hack. ... to avoid calling isValid recursively
                def validDirective():
                    if not self.values:
                        return True
                    if self.values[-1].strip().endswith(','):
                        return False
                    try:
                        compile(
                            'func(' + ''.join(self.values) + ')',
                            filename='<string>',
                            mode='eval')
                    except Exception:
                        return False
                    return True

                if validDirective() and self._action is not None:
                    return 'script'
                return 'directive'
            return 'statements'
        return None

    def isValid(self) -> bool:
        '''Determine if the statement, expression or directive is valid. Otherwise
        the parser will continue until a valid multi-line expression or statement
        can be found.'''
        if not self.values:
            return True
        try:
            if self.category() == 'directive':
                # we add func() because the expression can be multi-line and
                # can have keyword-argument like options
                #
                # However, python considers
                #
                #     func('value', )
                #
                # a valid syntax but we do want , to continue to the next line
                if self.values[-1].strip().endswith(','):
                    self.error_msg = 'Trailing ,'
                    return False
                # to allow type trait, we will have to test the expression as if in a function
                # definition, with something like "def func(a : str, b : list=[])"
                try:
                    compile(
                        'func(' + ''.join(self.values) + ')',
                        filename='<string>',
                        mode='eval')
                except Exception:
                    compile(
                        'def func(' + ''.join(self.values) + '):\n  pass',
                        filename='<string>',
                        mode='exec')
            elif self.category() == 'statements':
                compile((''.join(self.values)),
                        filename='<string>',
                        mode='exec')
            elif self.category() == 'script':
                #
                # A valid script has an identation defined at the first line. That is to say
                #
                # line 1
                # line 2
                #
                # is allowed
                #
                #     line 1
                #     line 2
                #
                # line 3
                #
                # is not so the addition of line 3 would fail. However, the last line
                # will be tested before inserted so this function will always return True
                return True
            else:
                raise RuntimeError(
                    f'Unrecognized expression type {self.category()}')
            return True
        except Exception as e:
            self.error_msg = repr(e)
            return False

    def empty(self) -> bool:
        '''If there is no content (comment does not count)'''
        return self.category() is None

    def extend(self, line: str) -> None:
        '''Extend the current directive, expression or script'''
        if self.category() == 'directive':
            self.add_directive(None, line)
        elif self.category() == 'script':
            self._script += line
        else:
            self.add_statement(line)

    def add_directive(self,
                      key: Optional[str],
                      value: str,
                      lineno: Optional[int] = None,
                      comment: str = '') -> None:
        '''Assignments are items with ':' type '''
        if key is None:
            # continuation of multi-line directive
            self.statements[-1][2] += value
            self.values.append(value)
            if self._action is not None:
                self._action_options += value
        else:
            # new directive, the comment before it are used
            self.statements.append([':', key, value, comment])
            self.values = [value]
        if lineno:
            self.lineno = lineno

    def add_script(self,
                   key: str,
                   value: str,
                   lineno: Optional[int] = None) -> None:
        '''script starts with key: value'''
        # we need a fake directive here because the : directive can be multi-line and
        # we need to borrow the syntax checking of directives here.
        self.statements.append([':', '__script__', ''])
        self.values = [value]
        self._action = key
        self._action_options = value
        if lineno:
            self.lineno = lineno

    def add_statement(self, line: str, lineno: Optional[int] = None) -> None:
        '''statements are regular python statements'''
        # there can be only one statement block
        if self.category() != 'statements':
            self.values = [line]
        else:
            self.values.append(line)
        if self.statements and self.statements[-1][0] == '!':
            self.statements[-1][-1] += line
        else:
            self.statements.append(['!', line])
        if lineno:
            self.lineno = lineno

    def wrap_script(self) -> None:
        '''convert action: script to task: action(script)'''
        if self._action is None:
            return
        # _action options can contain both runtime option and action options
        opt = self._action_options.strip()
        if self.statements[-1][0] != ':' or self.statements[-1][
                1] != '__script__':
            raise RuntimeError('Failed to parse script')
        # under window, the lines will be ended with \r\n, which will cause
        # trouble with textwrap.dedent.
        embedded_script = '\n'.join(self._script.splitlines()) + '\n'
        # dedent, which will not remove empty line, so _scrit will always ends with \n.
        self._script = textwrap.dedent(embedded_script)
        if self._script == embedded_script:
            first_words = self._script[:10].splitlines()[0]
            env.logger.warning(
                f'Embedding script "{first_words}..." without indentation is error-prone and will be deprecated in the future.'
            )
        # let us look for 'expand=""' in options
        if 'expand' in opt:
            sigil, opt = extract_option_from_arg_list(opt, 'expand', None)
            if sigil is None or sigil is False:
                # no expansion
                self._script = repr(self._script)
            elif sigil is True:
                self._script = as_fstring(self._script)
            else:
                self._script = as_fstring(replace_sigil(self._script, sigil))
        else:
            # verbatim script, use repr is enough
            self._script = repr(self._script)
        self.statements[-1] = [
            '!',
            f'{self._action}({self._script}{(", " + opt) if opt else ""})\n'
        ]
        self.values = []
        self._action = None
        self._action_options = None
        self._script = ''

    def get_tokens(self) -> str:
        '''Get tokens after input statement'''

        def _get_tokens(statement):
            return [
                x[1]
                for x in generate_tokens(StringIO(statement).readline)
                if x[1] not in ('', '\n')
            ]

        tokens: List = []
        for statement in self.statements:
            tokens.extend(
                _get_tokens(statement[2] if statement[0] ==
                            ':' else statement[1]))

        if self.task:
            tokens.extend(_get_tokens(self.task))

        return ' '.join(tokens)

    def finalize(self) -> None:
        ''' split statement and task by last directive '''
        self.wrap_script()
        if not self.statements:
            self.task = ''
            return
        # handle tasks
        input_directive = [
            idx for idx, statement in enumerate(self.statements)
            if statement[0] == ':' and statement[1] == 'input'
        ]
        task_directive = [
            idx for idx, statement in enumerate(self.statements)
            if statement[0] == ':' and statement[1] == 'task'
        ]
        if len(task_directive) > 1:
            raise ValueError('Only one task statement is allowed in a step')
        # handle parameter
        for idx, statement in enumerate(self.statements):
            if statement[0] == ':' and statement[1] == 'parameter':
                if task_directive and task_directive[0] < idx:
                    raise ValueError(
                        'Parameter statement is not allowed in tasks.')
                if '=' not in statement[2]:
                    if ':' in statement[2]:
                        if not is_type_hint(statement[2]):
                            raise ValueError(
                                f'Invalid type trait in parameter specification {statement[2]}'
                            )
                        name, value = statement[2].split(':')
                    else:
                        name = statement[2]
                        value = 'str'
                else:
                    name, value = statement[2].split('=', 1)
                    # ignore type trait if a default value is specified
                    name = name.split(':')[0]
                name = name.strip()
                if name.startswith('_'):
                    raise ValueError(
                        f'Invalid parameter name {name}: names with leading underscore is not allowed.'
                    )
                if not value.strip():
                    raise ValueError(
                        f'{self.step_name()}: Invalid parameter definition: {statement[2]}'
                    )
                # there is a possibility that value contains # so  sos_handle_parameter(name, val # aaa) will fail
                self.statements[idx] = [
                    '!',
                    f'#begin_parameter {name}\n{name} = sos_get_param({name.strip()!r}, {value}\n) #end_parameter {name}\n',
                    statement[2].strip()
                ]
                self.parameters[name] = (value, statement[3])
                if input_directive and input_directive[0] < idx:
                    self.substep_parameters.add(name)
        # handle tasks
        if not task_directive:
            self.task = ''
        else:
            start_task = task_directive[0] + 1
            # convert statement to task
            self.task = ''
            for statement in self.statements[start_task:]:
                if statement[0] == ':':
                    if statement[1] in ('input', 'output', 'depends'):
                        raise ValueError(
                            f'{self.step_name()}: Step task should be defined as the last item in a SoS step'
                        )
                    elif statement[1] == 'task':
                        raise ValueError(
                            f'{self.step_name()}: Only one task is allowed for a step'
                        )
                    elif statement[1] == 'parameter':
                        raise ValueError(
                            f'{self.step_name()}: Parameters should be defined before step task'
                        )
                    # ignore ...
                    self.task += '\n'
                else:
                    self.task += statement[1]
            self.task_params = self.statements[task_directive[0]][2]
            self.statements = self.statements[:task_directive[0]]
        # merge multiple statments at the end
        if len(self.statements) > 1 and self.statements[-1][0] == '!':
            starting = len(self.statements) - 1
            for idx in range(starting - 1, -1, -1):
                if self.statements[idx][0] == '!':
                    starting = idx
                else:
                    break
            # merge
            for idx in range(starting + 1, len(self.statements)):
                self.statements[starting][1] += self.statements[idx][1]
            # remove the rest of the statements
            self.statements = self.statements[:starting + 1]
        #
        # auto provides #859
        if not any(opt in self.options for opt in ('provides', 'shared')) and \
                len([x for x in self.statements if x[0] == ':' and x[1] == 'output']) == 1:
            output_stmt = [
                x for x in self.statements if x[0] == ':' and x[1] == 'output'
            ][0][2]
            output_names = get_names_of_kwargs(output_stmt)
            self.options['namedprovides'] = repr(output_names)

    def show(self):
        '''Output for command sos show'''
        textWidth = max(60, shutil.get_terminal_size((80, 20)).columns)
        text = f'  {self.step_name() + ":":<21} ' + self.comment
        print('\n'.join(
            textwrap.wrap(
                text,
                width=textWidth,
                initial_indent='',
                subsequent_indent=' ' * 24)))
        local_parameters = {
            x: y
            for x, y in self.parameters.items()
            if x not in self.global_parameters
        }
        if local_parameters:
            print('    Workflow Options:')
        for name, (value, comment) in local_parameters.items():
            par_str = f'      {format_par(name, value)}'
            print(par_str)
            if comment:
                print('\n'.join(
                    textwrap.wrap(
                        comment,
                        width=textWidth,
                        initial_indent=' ' * 24,
                        subsequent_indent=' ' * 24)))


class SoS_Workflow:
    '''A SoS workflow with multiple steps. It is created from multiple sections of a SoS script
    and consists of multiple SoS_Step.'''

    def __init__(self, content: 'SoS_ScriptContent', workflow_name: str,
                 allowed_steps: Optional[str], sections: List[SoS_Step],
                 global_stmts: str) -> None:
        '''create a workflow from its name and a list of SoS_Sections (using name matching)'''
        self.content = content
        self.name = workflow_name if workflow_name else 'default'
        self.sections: List = []
        self.auxiliary_sections: List = []
        self.global_stmts = global_stmts
        #
        for section in sections:
            for name, index, alias in section.names:
                self.auxiliary_sections.append(section)
                self.auxiliary_sections[-1].name = name
                self.auxiliary_sections[-1].index = index
                self.auxiliary_sections[-1].alias = alias
                self.auxiliary_sections[-1].uuid = uuid4()
                # an auxiliary step can also serve as a regular step
                # as long as it matches workflow name.
                if (name == '' and
                        workflow_name == 'default') or fnmatch.fnmatch(
                            workflow_name, name):
                    self.sections.append(copy.deepcopy(section))
                    self.sections[
                        -1].name = name if name == '' and workflow_name == 'default' else workflow_name
                    self.sections[-1].index = index
                    self.sections[-1].alias = alias
                    self.sections[-1].uuid = uuid4()
        #
        # sort sections by index
        self.sections.sort(key=lambda x: 0 if x.index is None else x.index)
        #
        # disable some disallowed steps
        if allowed_steps:
            all_steps = {
                x.index: False
                for x in self.sections
                if x.index and x.index >= 0
            }
            #
            for item in allowed_steps.split(','):
                # remove space
                item = ''.join([x for x in item if x != ' '])
                if item.isdigit():
                    # pipeline:100
                    all_steps[int(item)] = True
                elif '-' in item and item.count('-') == 1:
                    l, u = item.split('-')
                    if (l and not l.isdigit()) or (u and not u.isdigit()) or \
                            (l and u and int(l) > int(u)):
                        raise ValueError(f'Invalid pipeline step item {item}')
                    # pipeline:-100, pipeline:100+ or pipeline:10-100
                    if not l:
                        l = min(all_steps.keys())
                    if not u:
                        u = max(all_steps.keys())
                    #
                    for key in all_steps.keys():
                        if key >= int(l) and key <= int(u):
                            all_steps[key] = True
                else:
                    raise ValueError(f'Invalid pipeline step item {item}')
            # keep only selected steps (and the global section)
            self.sections = [
                x for x in self.sections if x.index < 0 or all_steps[x.index]
            ]
        # assign 'last_step' to each section
        last_step = None
        for section in self.sections:
            section.last_step = last_step
            last_step = section.step_name()
        # for auxiliary steps they may also have the step -1
        self.auxiliary_sections.sort(
            key=lambda x: (x.name, 0 if x.index is None else x.index))
        for idx, section in enumerate(self.auxiliary_sections):
            if idx == 0 or section.index is None:
                continue
            if self.auxiliary_sections[idx - 1].name == section.name:
                section.last_step = self.auxiliary_sections[idx - 1].step_name()

    def section_by_id(self, uuid: UUID) -> SoS_Step:
        for section in self.sections:
            if uuid == section.uuid:
                return section
        for section in self.auxiliary_sections:
            if uuid == section.uuid:
                return section
        raise RuntimeError(f'Failed to find section with uuid {uuid}')

    def extend(self, workflow: 'SoS_Workflow') -> None:
        '''Append another workflow to existing one to created a combined workflow'''
        # all sections are simply appended ...
        # but we will need to make sure that the new workflow is
        # executed after the previous one.
        if not workflow.sections:
            return
        if not self.sections:
            self.sections = workflow.sections
            return
        section = workflow.sections[0]
        depends_idx = [
            idx for idx, stmt in enumerate(section.statements)
            if stmt[0] == ':' and stmt[1] == 'depends'
        ]
        if not depends_idx:
            section.statements.insert(0, [
                ':', 'depends', f"sos_step('{self.sections[-1].step_name()}')"
            ])
        else:
            section.statements[depends_idx[0]][2] = section.statements[depends_idx[0]][2].strip() + \
                (", " if section.statements[depends_idx[0]][2].strip() else "") + \
                f"sos_step('{self.sections[-1].step_name()}')\n"
        self.sections.extend(workflow.sections)

    def has_external_task(self) -> bool:
        return any(x.has_external_task() for x in self.sections) or \
            any(x.has_external_task() for x in self.auxiliary_sections)

    def parameters(self) -> Dict[str, str]:
        # collect parameters defined by `parameter:` of steps
        par: Dict = {}
        for x in self.sections + self.auxiliary_sections:
            par.update(x.parameters)
        return {x: y[0] for x, y in par.items()}


class SoS_ScriptContent:
    '''A small class to record the script information to be used by nested
    workflow.'''

    def __init__(self,
                 content: str = '',
                 filename: Optional[str] = None) -> None:
        self.content = content
        self.filename = filename
        self.included = []
        self.md5 = self.calc_md5()

    def calc_md5(self) -> str:
        cnt = self.text()
        # additional files
        for script in self.included:
            cnt += script[0]
        #
        return textMD5(cnt)

    def text(self) -> str:
        if self.content:
            return self.content
        else:
            with open(self.filename) as script:
                return script.read()

    def __repr__(self):
        return f'{self.md5}: filename: {self.filename}, content: {self.content}'

    def __eq__(self, other):
        return self.md5 == other.md5

    def __ne__(self, other):
        return self.md5 != other.md5


class SoS_Script:

    def __init__(self,
                 content: Optional[str] = '',
                 filename: Optional[str] = None) -> None:
        '''Parse a sectioned SoS script file. Please refer to the SoS manual
        for detailed specification of this format.

        Parameter `content` can be either a filename or a content of a
        SoS script in unicode, which is convenient for passing scripts for
        testing purposes.

        Parameter `filename` should be used if the content should be read
        from a file.

        A SoS_Script generated from a .sos file contains the following items

        sections:
            A list of SoS_Step objects that present all sections read.
            Among other things, section.names
        '''
        if not content:
            try:
                content, self.sos_script = locate_script(filename, start='.')
            except Exception:
                # try to add .sos extension?
                if not filename.endswith('.sos'):
                    try:
                        content, self.sos_script = locate_script(
                            filename + '.sos', start='.')
                    except Exception:
                        if not filename.endswith('.ipynb'):
                            try:
                                content, self.sos_script = locate_script(
                                    filename + '.ipynb', start='.')
                            except Exception as e:
                                env.logger.debug(e)
                                env.logger.error(
                                    f'Failed to locate {filename}, {filename}.sos, or {filename}.ipynb'
                                )
                                sys.exit(1)
                        else:
                            raise
                else:
                    raise
            # Is this script in sos or ipynb format?
            ext = os.path.splitext(self.sos_script)[-1]
            if ext == '.ipynb':
                # convert ipynb to sos
                from .converter import extract_workflow
                content = extract_workflow(self.sos_script)
                self.sos_script = '<string>'
                self.content = SoS_ScriptContent(content, None)
            else:
                self.content = SoS_ScriptContent(content, self.sos_script)
        else:
            self.sos_script = '<string>'
            self.content = SoS_ScriptContent(content, None)
        # save a parsed version of the script for displaying purpose only
        self.global_stmts = ''

        self.description = []
        self._last_comment = ''
        # open the file
        if content:
            with StringIO(content) as fp:
                self._read(fp)
        else:
            with open(self.sos_script) as fp:
                self._read(fp)
        #
        # workflows in this script, from sections that are not skipped.
        all_section_steps = sum([x.names for x in self.sections], [])
        forward_section_steps = sum(
            [x.names for x in self.sections if 'provides' not in x.options], [])
        # (name, None) is auxiliary steps
        self.workflows = list(
            dict.fromkeys([x[0] for x in all_section_steps if '*' not in x[0]]))
        forward_workflows = list(
            dict.fromkeys(
                [x[0] for x in forward_section_steps if '*' not in x[0]]))
        if not forward_workflows:
            self.workflows.append('default')
            self.default_workflow = 'default'
        elif len(forward_workflows) == 1:
            self.default_workflow = forward_workflows[0]
        else:
            self.default_workflow = None
        #
        # now we need to record the workflows to the global section so
        # that we know if which has to be included when a subworkflow is used.
        for section in self.sections:
            if section.is_global:
                section.names = self.workflows

    def _find_include_file(self, sos_file: str) -> Tuple[str, str]:
        # we could almost use SoS_script directly but we need to be able to start searching
        # from path of the master file.
        try:
            if self.sos_script and self.sos_script != '<string>':
                start_path = os.path.split(self.sos_script)[0]
            else:
                start_path = ''
            try:
                content, script_file = locate_script(
                    sos_file + '.sos', start=start_path)
            except Exception:
                content, script_file = locate_script(
                    sos_file + '.ipynb', start=start_path)
                # convert ipynb to sos
                from .converter import extract_workflow
                content = extract_workflow(script_file)
        except Exception:
            raise RuntimeError(
                f'Source file for nested workflow {sos_file} with extension .sos or .ipynb does not exist'
            )

        return content, script_file

    def add_comment(self, line: str) -> None:
        '''Keeping track of "last comment" for section and parameter '''
        # the rule is like
        #
        # # comment line  --> add to last comment
        # blank line --> clears last comment
        # [ ] --> use last comment
        # parameter: --> use last comment
        # All others: clear last comment
        self._last_comment += (' ' if self._last_comment else '') + \
            line.lstrip('#').strip()

    def clear_comment(self):
        self._last_comment = ''

    def _read(self, fp: TextIOBase) -> None:
        self.sections: List = []
        self.format_version: str = '1.0'
        self.gloal_def: str = ''
        #
        comment_block = 1
        # cursect always point to the last section
        cursect = None
        all_step_names = []
        #
        # this ParsingError is a container for all parsing errors. It will be
        # raised after parsing if there is at least one parsing error.
        parsing_errors = ParsingError(self.sos_script)
        for lineno, line in enumerate(fp, start=1):
            #
            # for structural lines
            if SOS_MAGIC.match(line):
                continue

            if SOS_CELL.match(line):
                continue

            # comments in SoS scripts are mostly informative
            if line.startswith('#'):
                # Comment blocks before any section
                self.add_comment(line)
                if cursect is None:
                    if comment_block == 1:
                        # look for format information
                        mo = SOS_FORMAT_LINE.match(line)
                        if mo:
                            format_name = mo.group('format_name')
                            if not format_name.upper().startswith('SOS'):
                                parsing_errors.append(
                                    lineno, line,
                                    f'Unrecognized file format name {format_name}. Expecting SOS.'
                                )
                            mo = SOS_FORMAT_VERSION.match(format_name)
                            if mo:
                                self.format_version = mo.group('format_version')
                            else:
                                parsing_errors.append(
                                    lineno, line,
                                    f'Unrecognized file format version in {format_name}.'
                                )
                    elif comment_block == 2:
                        # anything before the first section can be pipeline
                        # description.
                        if cursect is None:
                            self.description.append(line)
                else:
                    # this is description of the section
                    if cursect.empty():
                        pass
                    # this is comment in scripts (and perhaps not even comment)
                    elif cursect.category() == 'script':
                        if cursect.indented_script():
                            # if the script is indented and encounters a comment
                            # from first column, switch to comment mode
                            try:
                                cursect.wrap_script()
                            except Exception as e:
                                parsing_errors.append(lineno, line, str(e))
                        else:
                            cursect.extend(line)
                    # this can be comment or back comment
                    elif cursect.category() == 'statements':
                        if not cursect.isValid():
                            cursect.extend(line)
                continue
            elif not line.strip():
                # a blank line start a new comment block if we are still
                # in the front of the script
                self.clear_comment()
                if cursect is None:
                    comment_block += 1
                else:
                    category = cursect.category()
                    if category in ('statements', 'script'):
                        cursect.extend(line)
                    elif category == 'directive' and not cursect.isValid():
                        # 1311
                        parsing_errors.append(
                            lineno, line,
                            f'Ending incomplete options ({"".join(cursect.values).strip()}) with an empty line'
                        )
                continue

            #
            # a continuation of previous item?
            if line[0].isspace(
            ) and cursect is not None and not cursect.empty():
                mo = SOS_INDENTED_ACTION.match(line)

                if mo:
                    #
                    # case 0:
                    #
                    # [1]
                    #    python: whatever
                    #
                    if not cursect:
                        parsing_errors.append(
                            cursect.lineno, ''.join(cursect.values[:5]),
                            'Section cannot start from indented action.')
                        continue

                    # case 1:
                    #
                    # [1]
                    # if True:
                    #    report:
                    #        something
                    #    whatever: <- this line
                    if cursect.category() == 'script':
                        if cursect.indented_script() > re.search(r'\S',
                                                                 line).start():
                            try:
                                cursect.wrap_script()
                            except Exception as e:
                                parsing_errors.append(
                                    cursect.lineno, ''.join(cursect.values[:5]),
                                    str(e))
                        else:
                            # case 2:
                            #
                            # if True:
                            #     report:
                            #        name: whatever
                            cursect.extend(line)
                            continue

                    # not script, or a the script has been wrapped
                    action_name = mo.group('action_name')
                    # newline should be kept in case of multi-line directive
                    action_value = mo.group('action_value') + '\n'
                    cursect.add_script(action_name, action_value, lineno)
                elif cursect.indented_script() > re.search(r'\S', line).start():
                    # case of wrapping previous script with NO indented action
                    #
                    # if True:
                    #     sh:
                    #        has indent
                    #     this line <-  NOT match
                    # or this line <- NOT match
                    try:
                        cursect.wrap_script()
                    except Exception as e:
                        parsing_errors.append(lineno, line, str(e))
                    cursect.extend(line)
                else:
                    # other cases
                    #
                    cursect.extend(line)
                continue
            #
            # is it a continuation of uncompleted directive?
            if cursect and not cursect.isValid():
                cursect.extend(line)
                continue
            #
            # a new line (start from first column)
            #
            # section header?
            mo = SOS_SECTION_HEADER.match(line)
            if mo:
                # check previous expression before a new section
                if cursect:
                    if not cursect.isValid():
                        parsing_errors.append(
                            cursect.lineno, ''.join(cursect.values[:5]),
                            f'Invalid {cursect.category()}: {cursect.error_msg}'
                        )
                    cursect.values = []
                    try:
                        cursect.finalize()
                    except Exception as e:
                        parsing_errors.append(cursect.lineno,
                                              ''.join(cursect.values[:5]),
                                              str(e))
                # start a new section
                section_name = mo.group('section_name').strip()
                section_option = mo.group('section_option')
                step_names = []
                step_options = {}
                #
                for name in section_name.split(','):
                    mo = SOS_SECTION_NAME.match(name)
                    if mo:
                        n, i, di, al = mo.group('name', 'index',
                                                'default_index', 'alias')
                        if n == 'global' and i is not None:
                            parsing_errors.append(
                                lineno, line,
                                'Invalid global section definition')
                        if n == '':
                            parsing_errors.append(
                                lineno, line, 'Empty step name is not allowed')
                        if i and str(int(i)) != i:
                            # disallow cases such as a_01
                            parsing_errors.append(
                                lineno, line,
                                f'Invalid section index {i} (leading zero is not allowed)'
                            )
                        if n:
                            if i is None and '*' in n:
                                parsing_errors.append(
                                    lineno, line,
                                    'Unindexed section name cannot contain wildcard character (*).'
                                )
                            step_names.append([n, int(i) if i else i, al])
                        if di:
                            step_names.append(['', int(di), al])
                    else:
                        parsing_errors.append(lineno, line,
                                              'Invalid section name')
                if 'global' in [x[0] for x in step_names
                               ] and len(step_names) > 1:
                    parsing_errors.append(
                        lineno, line,
                        'Global section cannot be shared with another step')
                if section_option is not None:
                    # this does not work directly because list parameter can have ,
                    # without having to really evaluate all complex expressions, we
                    # have to try to put syntax correctly pieces together.
                    try:
                        pieces = separate_options(section_option)
                        for option in pieces:
                            mo = SOS_SECTION_OPTION.match(option)
                            if mo:
                                opt_name, opt_value = mo.group('name', 'value')
                                if opt_name in step_options:
                                    parsing_errors.append(
                                        lineno, line, 'Duplicate options')
                                step_options[opt_name] = opt_value
                            else:
                                parsing_errors.append(lineno, line,
                                                      'Invalid section option')
                    except Exception as e:
                        parsing_errors.append(lineno, line, str(e))
                    if 'EXECUTOR' in env.config[
                            'SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
                        env.log_to_file(
                            'EXECUTOR',
                            'Header parsed with names {} and options {}'.format(
                                step_names, step_options))
                for name in step_names:
                    prev_workflows = [
                        x[0] for x in all_step_names if '*' not in x[0]
                    ]
                    for prev_name in all_step_names:
                        # auxiliary step
                        if name[1] is None and prev_name[1] is None and name[
                                0] != prev_name[0]:
                            continue
                        # index not euqal (one of them can be None)
                        if name[1] != prev_name[1]:
                            continue
                        # index equal and one of them have wild card character
                        if '*' in name[0]:
                            names = [
                                x for x in prev_workflows
                                if re.match(name[0].replace('*', '.*'), x)
                            ]
                        else:
                            names = [name[0]]
                        if '*' in prev_name:
                            prev_names = [
                                x for x in prev_workflows
                                if re.match(prev_name[0].replace('*', '.*'), x)
                            ]
                        else:
                            prev_names = [prev_name[0]]
                        if len(set(prev_names)
                               & set(names)) and 'global' not in names:
                            parsing_errors.append(
                                lineno, line, f'Duplicate section name {names}')
                all_step_names.extend(step_names)
                if 'global' in [x[0] for x in step_names]:
                    if step_options:
                        parsing_errors.append(
                            lineno, line,
                            'Global section does not accept any option')
                    self.sections.append(SoS_Step(is_global=True))
                else:
                    # the second block is attached to
                    if comment_block == 2:
                        self.description = ''
                    self.sections.append(
                        SoS_Step(
                            self.content,
                            step_names,
                            step_options,
                            comment=self._last_comment))
                cursect = self.sections[-1]
                self.clear_comment()
                continue
            #
            # directive?
            mo = SOS_DIRECTIVE.match(line)
            if mo and not is_type_hint(line):
                # check previous expression before a new directive
                if cursect:
                    if not cursect.isValid():
                        parsing_errors.append(
                            cursect.lineno, ''.join(cursect.values[:5]),
                            f'Invalid {cursect.category()}: {cursect.error_msg}'
                        )
                    cursect.values = []
                    # allow multiple process-style actions
                    try:
                        cursect.wrap_script()
                    except Exception as e:
                        parsing_errors.append(cursect.lineno,
                                              ''.join(cursect.values[:5]),
                                              str(e))
                else:
                    self.sections.append(SoS_Step(is_global=True))
                    cursect = self.sections[-1]
                #
                directive_name = mo.group('directive_name')
                # newline should be kept in case of multi-line directive
                directive_value = mo.group('directive_value') + '\n'
                # is it an action??
                if directive_name in SOS_DIRECTIVES and directive_name != 'parameter':
                    if cursect is None:
                        parsing_errors.append(
                            lineno, line,
                            f'Directive {directive_name} is not allowed outside of a SoS step'
                        )
                        continue
                    cursect.add_directive(directive_name, directive_value,
                                          lineno)
                else:
                    if directive_name == 'parameter':
                        if not directive_value.strip():
                            # ignore lines with only an empty parameter directive #1282
                            continue
                        if comment_block == 2:
                            self.description = ''
                        cursect.add_directive(
                            directive_name,
                            directive_value,
                            lineno,
                            comment=self._last_comment)
                    else:
                        # let us check if this is an acture action, or a type hint
                        cursect.add_script(directive_name, directive_value,
                                           lineno)
                self.clear_comment()
                continue
            # if section is in script mode?
            if cursect and cursect.isValid() and cursect.category() == 'script':
                # if the script is indented and the line is not, the script
                # is ended.
                if not line[0].isspace() and cursect.indented_script():
                    try:
                        cursect.wrap_script()
                    except Exception as e:
                        parsing_errors.append(lineno, line, str(e))
                else:
                    cursect.extend(line)
                    continue

            # all others?
            if not cursect:
                self.sections.append(SoS_Step(is_global=True))
                cursect = self.sections[-1]
                cursect.add_statement(line, lineno)
                self.clear_comment()
                continue
            #
            if cursect.empty() or cursect.category() != 'statements':
                # new statement
                cursect.add_statement(line, lineno)
            else:
                # existing one
                cursect.extend(line)
        #
        # check the last expression before a new directive
        if cursect:
            if not cursect.isValid():
                parsing_errors.append(
                    cursect.lineno, ''.join(cursect.values[:5]),
                    f'Invalid {cursect.category()}: {cursect.error_msg}')
            else:
                try:
                    cursect.finalize()
                except Exception as e:
                    parsing_errors.append(cursect.lineno,
                                          ''.join(cursect.values[:5]), str(e))

        #
        # if there is any parsing error, raise an exception
        #
        # if there is no section in the script, we create a default section with global
        # definition being the content.
        global_parameters: Dict = {}
        if not [x for x in self.sections if not x.is_global]:
            self.sections.append(
                SoS_Step(self.content, [('default', None, None)]))
            for section in [x for x in self.sections if x.is_global]:
                if self.sections[-1].task != '':
                    parsing_errors.append(
                        cursect.lineno, 'Invalid section',
                        'Cannot define multiple default sections with a task in between.'
                    )
                self.sections[-1].statements.extend(section.statements)
                self.sections[-1].task = section.task
                self.sections[-1].task_params = section.task_params
                self.global_stmts = ''
                global_parameters.update(section.parameters)
            # The sections should have been finalized so there is no need to finalize
            # again. In particular, finalizing a section would reset existing task #833
            # self.sections[-1].finalize()
        else:
            # as the last step, let us insert the global section to all sections
            for sec in [x for x in self.sections if x.is_global]:
                for statement in sec.statements:
                    if statement[0] == ':':
                        parsing_errors.append(
                            cursect.lineno, f'{statement[1]}:{statement[2]}',
                            'Global section cannot contain sos input, ouput, and task statements'
                        )
                    else:
                        self.global_stmts += statement[1]
                global_parameters.update(sec.parameters)
        # remove the global section after inserting it to each step of the process
        self.sections = [x for x in self.sections if not x.is_global]
        #
        if parsing_errors.errors:
            raise parsing_errors
        #
        for section in self.sections:
            section.global_stmts = self.global_stmts
            section.global_parameters = global_parameters
            section.parameters.update(global_parameters)
            #
            section.md5 = textMD5(section.get_tokens())

    def workflow(self,
                 workflow_name: Optional[str] = None,
                 use_default: bool = True) -> SoS_Workflow:
        '''Return a workflow with name_step+name_step specified in wf_name
        This function might be called recursively because of nested
        workflow.'''
        if workflow_name is None and not use_default:
            return SoS_Workflow(self.content, '', '', self.sections,
                                self.global_stmts)
        allowed_steps = None
        if not workflow_name:
            wf_name = ''
        else:
            # if consists of multiple workflows
            if '+' in workflow_name:
                wfs = []
                for wf in workflow_name.split('+'):
                    if not SOS_SUBWORKFLOW.match(wf):
                        raise ValueError(
                            f'Incorrect workflow name {workflow_name}')
                    # if this is a combined workflow, extra_section might be specied.
                    wfs.append(self.workflow(wf))
                combined_wf = wfs[0]
                for wf in wfs[1:]:
                    combined_wf.extend(wf)
                combined_wf.name = workflow_name
                return combined_wf
            # if a single workflow
            # workflow_10:15 etc
            mo = SOS_SUBWORKFLOW.match(workflow_name)
            if not mo:
                raise ValueError(f'Incorrect workflow name {workflow_name}')
            wf_name, allowed_steps = mo.group('name', 'steps')
        # check source
        if not wf_name:
            if len(self.workflows) == 1:
                wf_name = list(self.workflows)[0]
            elif self.default_workflow:
                wf_name = self.default_workflow
            elif 'default' in self.workflows or '' in self.workflows:
                wf_name = 'default'
            else:
                raise ValueError(
                    'Name of workflow should be specified because '
                    'the script defines more than one pipelines without a default one. '
                    'Available pipelines are: {}.'.format(', '.join(
                        self.workflows)))
        elif wf_name not in self.workflows and wf_name != 'default':
            raise ValueError(
                f'Workflow {wf_name} is undefined. Available workflows are: {", ".join(self.workflows)}'
            )

        return SoS_Workflow(self.content, wf_name, allowed_steps, self.sections,
                            self.global_stmts)

    def print_help(self, script_name: str):
        '''print a help message from the script'''
        textWidth = max(60, shutil.get_terminal_size((80, 20)).columns)

        if len(script_name) > 20:
            print(f'usage: sos run {script_name}')
            print(
                '               [workflow_name | -t targets] [options] [workflow_options]'
            )
        else:
            print(
                f'usage: sos run {script_name} [workflow_name | -t targets] [options] [workflow_options]'
            )
        print(
            '  workflow_name:        Single or combined workflows defined in this script'
        )
        print('  targets:              One or more targets to generate')
        print(
            '  options:              Single-hyphen sos parameters (see "sos run -h" for details)'
        )
        print(
            '  workflow_options:     Double-hyphen workflow-specific parameters'
        )
        description = [x.lstrip('# ').strip() for x in self.description]
        description = textwrap.dedent('\n'.join(description)).strip()
        if description:
            print('\n' + description)
        #
        print('\nWorkflows:')
        print('  ' + '\n  '.join(self.workflows))
        #
        global_parameters = {}
        for section in self.sections:
            global_parameters.update(section.global_parameters)
        if global_parameters:
            print('\nGlobal Workflow Options:')
            for name, (value, comment) in global_parameters.items():
                par_str = f'  {format_par(name, value)}'
                print(par_str)
                if comment:
                    print('\n'.join(
                        textwrap.wrap(
                            comment,
                            width=textWidth,
                            initial_indent=' ' * 24,
                            subsequent_indent=' ' * 24)))
        #
        print('\nSections')
        for section in self.sections:
            section.show()



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import glob
import os
import shutil
import socket
import sys
import time
import unittest
import pytest

from sos import execute_workflow

from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env
from sos.workflow_executor import Base_Executor
from sos import execute_workflow


def internet_on(host='8.8.8.8', port=80, timeout=3):
    '''Test if internet is connected '''
    try:
        socket.setdefaulttimeout(timeout)
        socket.create_connection(("www.google.com", 80))
        return True
    except Exception as e:
        print(e)
        return False


with_network = internet_on()


def testAcceptableArgs():
    '''test acceptable args of options'''
    with pytest.raises(Exception):
        execute_workflow(r"""
            run: unrecog=1
            echo 'a'
            """)


def testGetOutput():
    '''Test utility function get_output'''
    execute_workflow(r"""
        [0: shared='ret']
        ret = get_output('echo blah')
        """)
    # use strip because there would be \r\n under windows
    assert env.sos_dict['ret'].strip() == 'blah'

    execute_workflow(r"""
        [0: shared='ret']
        ret = get_output('echo blah', show_command=True)
        """)
    assert [x.strip() for x in env.sos_dict['ret'].splitlines()
           ] == ['$ echo blah', 'blah']

    execute_workflow(r"""
        [0: shared='ret']
        ret = get_output('echo blah', show_command=True, prompt='% ')
        """)
    assert [x.strip() for x in env.sos_dict['ret'].splitlines()
           ] == ['% echo blah', 'blah']

    with pytest.raises(Exception):
        execute_workflow(r"""
            [0]
            get_output('catmouse')
        """)

    with pytest.raises(Exception):
        execute_workflow(r"""
            [0]
            ret = get_output('cat -h')
            """)


def testFailIf(temp_factory):
    '''Test action fail if'''
    temp_factory('a.txt')

    # should fail in dryrun mode
    with pytest.raises(Exception):
        execute_workflow(r"""
            [0]
            input: 'a.txt'
            fail_if(len(input) == 1)
            """)

    with pytest.raises(Exception):
        execute_workflow(r"""
            [0]
            input: 'a.txt', 'b.txt'
            fail_if(len(input) == 2)
        """)


def testDelayedFailIf():
    # test fail_if of killing another running substep
    st = time.time()
    with pytest.raises(Exception):
        execute_workflow(
            r"""
            import time

            [10]
            time.sleep(8)

            [20]
            input: None
            time.sleep(2)
            fail_if(True)
            """,
            config={'worker_procs': ['3']})

    assert time.time(
    ) - st >= 8, 'Test test should fail only after step 10 is completed'


def testDelayedFailIfFromNestedWorkflow():
    # test fail_if of killing another running substep

    st = time.time()
    with pytest.raises(Exception):
        execute_workflow(
            r"""
            import time

            [default]
            sos_run('a')

            [a_10]
            time.sleep(8)

            [a_20]
            input: None
            time.sleep(2)
            fail_if(True)
        """,
            config={'worker_procs': ['3']})

    assert time.time(
    ) - st >= 8, 'Test test should fail only after step 10 is completed'


def testWarnIf(temp_factory):
    '''Test action fail if'''
    execute_workflow(
        r"""
        [0]
        warn_if(input is None, 'Expect to see a warning message')
        """,
        options={'run_mode': 'dryrun'})

    temp_factory('a.txt', 'b.txt')
    execute_workflow(
        r"""
        [0]
        input: 'a.txt', 'b.txt'
        warn_if(len(_input) == 1)
        """,
        options={'run_mode': 'dryrun'})


def testStopIf():
    '''Test action stop_if'''
    execute_workflow(r'''
        [0: shared='result']
        rep = range(20)
        result = []
        input: for_each='rep'

        stop_if(_rep > 10)
        result.append(_rep)
        ''')
    assert env.sos_dict['result'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


def test_stop_if_1(clear_now_and_after):
    # stop_if should not be treated as error so the previously
    # generated output file will be removed
    clear_now_and_after('test_stop_if_0.txt', 'test_stop_if_1.txt')

    execute_workflow(r'''
        [10]
        rep = range(2)
        input: for_each='rep'
        output: f'test_stop_if_{_rep}.txt'

        _output.touch()
        stop_if(_rep == 1, no_output=True)

        [20]
        assert(step_input.contains('test_stop_if_0.txt'))
        assert(not step_input.contains('test_stop_if_1.txt'))
        ''')

    assert os.path.isfile('test_stop_if_0.txt')
    assert not (os.path.isfile('test_stop_if_1.txt'))


def test_stop_if_2(clear_now_and_after):
    # stop_if should not be treated as error so the previously
    # generated output file will not be removed
    clear_now_and_after('test_stop_if_0.txt', 'test_stop_if_1.txt')

    execute_workflow(r'''
        [10]
        rep = range(2)
        input: for_each='rep'
        output: f'test_stop_if_{_rep}.txt'

        _output.touch()
        stop_if(_rep == 1)

        [20]
        assert(step_input.contains('test_stop_if_0.txt'))
        assert(step_input.contains('test_stop_if_1.txt'))
        ''')
    assert os.path.isfile('test_stop_if_0.txt')
    assert os.path.isfile('test_stop_if_1.txt')


def testSkipIf():
    '''Test action stop_if'''
    execute_workflow(r'''
        [0: shared='result']
        rep = range(20)
        result = []
        input: for_each='rep'

        skip_if(_rep > 10)
        result.append(_rep)
        ''')
    assert env.sos_dict['result'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


def test_skip_if_1(clear_now_and_after):
    clear_now_and_after([f'test_stop_if_{rep}.txt' for rep in range(2)])

    # stop_if should not be treated as error so the previously
    # generated output file will be removed

    execute_workflow(r'''
        [10]
        rep = range(2)
        input: for_each='rep'
        output: f'test_stop_if_{_rep}.txt'

        _output.touch()
        skip_if(_rep == 1)

        [20]
        assert(step_input.contains('test_stop_if_0.txt'))
        assert(not step_input.contains('test_stop_if_1.txt'))

        ''')

    assert os.path.isfile('test_stop_if_0.txt')
    assert not os.path.isfile('test_stop_if_1.txt')


def testDoneIf(clear_now_and_after):
    'Test action done_if'
    clear_now_and_after([f'test_done_if_{rep}.txt' for rep in range(2)])

    execute_workflow(r'''
        [10]
        rep = range(2)
        input: for_each='rep'
        output: f'test_done_if_{_rep}.txt'

        _output.touch()
        done_if(_rep == 1)

        [20]
        assert(step_input.contains('test_done_if_0.txt'))
        assert(step_input.contains('test_done_if_1.txt'))

        ''')
    assert (os.path.isfile('test_done_if_0.txt'))
    assert (os.path.isfile('test_done_if_1.txt'))


def testRun():
    '''Test action run'''
    execute_workflow(r'''
        [0]
        run:
        echo 'Echo'
        ''')


@pytest.mark.skipif(
    sys.platform == 'win32', reason="Under windows, echo 'Echo is perfectly OK")
def test_run_1():

    with pytest.raises(Exception):
        execute_workflow(r'''
            [0]
            run:
            echo 'Echo
            ''')


def testRunWithShebang():
    execute_workflow(r'''
        [0]
        run:
        #!/usr/bin/env python
        print('Echo')
        ''')


@pytest.mark.skipif(not shutil.which('perl'), reason="Needs perl")
def testPerl():
    '''Test action perl'''
    execute_workflow(r'''
        [0]
        perl:
        use strict;
        use warnings;

        print "hi NAME\n";
        ''')


@pytest.mark.skipif(not shutil.which('ruby'), reason="Needs ruby")
def testRuby():
    '''Test action ruby'''
    execute_workflow(r'''
        [0]
        ruby:
        line1 = "Cats are smarter than dogs";
        line2 = "Dogs also like meat";

        if ( line1 =~ /Cats(.*)/ )
        puts "Line1 contains Cats"
        end
        if ( line2 =~ /Cats(.*)/ )
        puts "Line2 contains  Dogs"
        end
        ''')


@pytest.mark.skipif(
    not with_network or 'TRAVIS' in os.environ,
    reason='Skip test because of no internet connection or in travis test')
def testDownload(temp_factory, clear_now_and_after):
    '''Test download of resources'''

    clear_now_and_after('tmp')
    temp_factory(dir='tmp')

    # test decompress tar.gz file
    execute_workflow(r'''
        [0]
        download(['http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus-20170912.DB.gz'],
        dest_dir='tmp', decompress=True)
        ''')
    assert (os.path.isfile('tmp/CancerGeneCensus-20170912.DB'))

    #
    # testing the download of single file
    #
    execute_workflow(r'''
        [0]
        download: dest_file='tmp/refgene.ppp'
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/resource/refgene.pkl
        ''')

    assert (os.path.isfile('tmp/refgene.ppp'))

    # test option dest_dir
    execute_workflow(r'''
        [0]
        download: dest_dir='tmp'
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/resource/refgene.pkl
        ''')
    assert (os.path.isfile('tmp/refgene.pkl'))


@pytest.mark.skipif(
    not with_network or 'TRAVIS' in os.environ,
    reason='Skip test because of no internet connection or in travis test')
def testDownloadMissingFile(temp_factory, clear_now_and_after):

    clear_now_and_after('tmp')
    temp_factory(dir='tmp')

    with pytest.raises(Exception):
        execute_workflow(r'''
            [0]
            download: dest_dir='tmp', decompress=True, max_jobs=2
            http://bioinformatics.mdanderson.org/Software/VariantTools/repository/resource/non-existing.gz
            http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus-20170912.DB.gz
            http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus.ann
            http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/DGV-hg38_20160831.ann
            ''')

    assert (os.path.isfile('tmp/CancerGeneCensus.ann'))


@pytest.mark.skipif(
    not with_network or 'TRAVIS' in os.environ,
    reason='Skip test because of no internet connection or in travis test')
def testDownloadLargeFile(temp_factory, clear_now_and_after):
    # test decompress tar.gz, .zip and .gz files

    clear_now_and_after('tmp')
    temp_factory(dir='tmp')

    execute_workflow(r'''
        [0]
        download: dest_dir='tmp', decompress=True
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus-20170912.DB.gz
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus.ann
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/DGV-hg38_20160831.ann
        ''')

    # run in build mode
    execute_workflow(r'''
        [0]
        download: dest_dir='tmp', decompress=True
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus-20170912.DB.gz
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus.ann
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/DGV-hg38_20160831.ann
        ''')


@pytest.mark.skipif(not shutil.which('pandoc'), reason="Needs pandoc")
def testPandoc(clear_now_and_after):
    '''Test action pandoc'''
    clear_now_and_after('report.md', 'myreport.html')

    execute_workflow(r'''
        [10]

        report: output='report.md'
        ## Some random figure

        Generated by matplotlib


        [100]
        # generate report
        output: 'myreport.html'
        pandoc(input='report.md', output=_output[0])
        ''')

    assert (os.path.isfile('myreport.html'))


@pytest.mark.skipif(not shutil.which('pandoc'), reason="Needs pandoc")
def test_pandoc_1(clear_now_and_after):
    clear_now_and_after('a.md', 'myreport.html')

    execute_workflow(r'''
        [10]
        report: output='a.md'
        ## Some random figure

        Generated by matplotlib


        [100]
        # generate report
        output: 'myreport.html'
        pandoc(input='a.md', output=_output[0])
        ''')

    assert os.path.isfile('myreport.html')


@pytest.mark.skipif(not shutil.which('pandoc'), reason="Needs pandoc")
def test_pandoc_2(clear_now_and_after):
    clear_now_and_after('a.md', 'myreport.html')
    #
    # another case is no output
    execute_workflow(r'''
        [10]
        report: output='a.md'
        ## Some random figure

        Generated by matplotlib


        [100]
        # generate report
        pandoc(input='a.md')
        ''')


@pytest.mark.skipif(not shutil.which('pandoc'), reason="Needs pandoc")
def test_pandoc_2(clear_now_and_after):
    # test acceptance of a list of input filenames
    clear_now_and_after('default_10.md', 'default_20.md', 'output.html')

    execute_workflow(r'''
        [10]
        report: output='default_10.md'
        A_10

        [20]
        report: output='default_20.md'
        A_20

        [100]
        # generate report
        pandoc(input=['default_10.md', 'default_20.md'], output='output.html')
        ''')

    for f in ['default_10.md', 'default_20.md', 'output.html']:
        assert file_target(f).target_exists()


def testReport(clear_now_and_after):
    '''Test action report'''
    clear_now_and_after('report.txt')

    script = r'''
        [A]
        parameter: num=5
        report: output='report.txt', expand=True
        touch {num}.txt

        '''
    # run twice
    execute_workflow(script, args=['--num', '7'])
    execute_workflow(script, args=['--num', '5'])

    with open('report.txt') as report:
        assert report.read() == 'touch 5.txt\n\n'


def test_report_1(clear_now_and_after):
    '''Test action report'''
    clear_now_and_after('report.txt')

    execute_workflow(r'''
        [A]
        report: output='report.txt', expand=True
        {step_name}

        [A_10]
        report: output='report.txt', expand=True
        {step_name}
        ''')
    with open('report.txt') as report:
        assert report.read() == 'A_10\n\n'


def test_report_2(clear_now_and_after):
    '''Test action report'''
    clear_now_and_after('a.txt', 'out.txt')

    execute_workflow(r'''
        [A_1]
        run: output='a.txt'
            echo something > a.txt

        report(input='a.txt', output='out.txt')
        ''')

    for name in ('a.txt', 'out.txt'):
        with open(name) as report:
            assert report.read().strip() == 'something'


def test_report_3(clear_now_and_after):
    '''Test action report'''
    clear_now_and_after('report.txt')
    #
    execute_workflow(r'''
        [A_1]
        run: output='a.txt'
        echo something > a.txt

        [A_2]
        run: output='b.txt'
        echo something else > b.txt

        [A_3]
        report(input=['a.txt', 'b.txt'], output='out.txt')
        ''')
    for name in ('a.txt', 'b.txt', 'out.txt'):
        assert (file_target(name).target_exists())


def test_report_4(clear_now_and_after):
    '''Test action report'''
    clear_now_and_after('a.txt')

    # test report to other types of output: path
    execute_workflow(r'''
        [A_1]
        report: output=path('a.txt')
        something
        ''')


def test_report_5(clear_now_and_after):
    '''Test action report'''
    clear_now_and_after('a.txt')

    execute_workflow(r'''
        [A_1]
        output: 'a.txt'
        report: output=_output[0]
        something
        ''')


def test_report_6(clear_now_and_after):
    '''Test action report'''
    clear_now_and_after('a.txt')

    # test report to other types of output: sos_targets
    execute_workflow(r'''
        [A_1]
        output: 'a.txt'
        report: output=_output
        something
        ''')


def test_report_7(clear_now_and_after):
    '''Test action report'''
    clear_now_and_after('a.txt')

    with pytest.raises(Exception):
        execute_workflow(r'''
            [A_1]
            output: 'a.txt', 'b.txt'
            report: output=_output
            something
            ''')


def testOptionWorkdir(temp_factory):
    '''Test option workdir of tasks'''
    temp_factory(dir='temp_wdr')

    with open(os.path.join('temp_wdr', 'a.txt'), 'w') as tmp:
        tmp.write('hello')

    execute_workflow(r'''
        [A_1]
        run: workdir='temp_wdr'
        cp -f a.txt a2.txt
        ''')

    assert file_target(os.path.join('temp_wdr', 'a2.txt')).target_exists()
    with open(os.path.join('temp_wdr', 'a.txt')) as tmp:
        assert 'hello' == tmp.read()


def testActionScript():
    '''Test action script'''
    execute_workflow(r'''
        [A_1]
        script: interpreter='python'
        with open('something.txt', 'w') as tmp:
            tmp.write('something')
        ''')

    assert file_target('something.txt').target_exists()
    with open('something.txt') as tmp:
        assert 'something' == tmp.read()


def testRegenerateReport(clear_now_and_after):
    '''Testing the regeneration of report once is needed. The problem
    here is the 'input' parameter of report.'''
    clear_now_and_after('a1.md', 'a2.md', 'out.md')
    script = r'''
        [A_1]
        output: 'a1.txt', 'a1.md'
        run:
        echo 'a1' >> a1.txt

        report: output='a1.md'
        a1

        [A_2]
        output: 'a2.txt', 'a2.md'
        run:
        echo 'a2' >> a2.txt
        report: output='a2.md'
        a2

        [A_3]
        input: 'a1.md', 'a2.md'
        output: 'out.md'
        report:     input=['a1.md', 'a2.md'], output='out.md'
        '''
    execute_workflow(script)
    with open('a1.md') as a:
        assert a.read() == 'a1\n\n'
    with open('a2.md') as a:
        assert a.read() == 'a2\n\n'
    with open('out.md') as a:
        assert a.read() == 'a1\n\na2\n\n'

    clear_now_and_after('a1.md', 'a2.md', 'out.md')
    execute_workflow(script)


def testActiveActionOption(temp_factory):
    '''Test the active option of actions'''
    # disallow
    with pytest.raises(Exception):
        SoS_Script('''
[1]
rep = range(5)
input: for_each = 'rep'
# ff should change and be usable inside run
ff = f"{_rep}.txt"
run:  expand=True, active=1,2
echo {ff}
touch temp/{ff}
''')
    #
    for active, result in [
        ('0', ['temp/0.txt']),
        ('-1', ['temp/4.txt']),
        ('(1,2)', ['temp/1.txt', 'temp/2.txt']),
        ('[2,3]', ['temp/2.txt', 'temp/3.txt']),
        ('(0,2,4)', ['temp/0.txt', 'temp/2.txt', 'temp/4.txt']),
        ('slice(1,None)',
         ['temp/1.txt', 'temp/2.txt', 'temp/3.txt', 'temp/4.txt']),
        ('slice(1,-2)', ['temp/1.txt', 'temp/2.txt']),
        ('slice(None,None,2)', ['temp/0.txt', 'temp/2.txt', 'temp/4.txt']),
        ('True',
         ['temp/0.txt', 'temp/1.txt', 'temp/2.txt', 'temp/3.txt',
          'temp/4.txt']),
        ('False', []),
    ]:
        temp_factory(dir='temp')
        # test first iteration
        execute_workflow(
            ('''
            [1]
            rep = range(5)
            input: for_each = 'rep'
            # ff should change and be usable inside run
            ff = f"{_rep}.txt"
            run:  expand=True, active=%s
            echo {ff}
            touch temp/{ff}
            ''' % active).replace('/', os.sep),
            options={'sig_mode': 'force'})

        files = list(glob.glob(os.path.join('temp', '*.txt')))
        assert sorted(files) == sorted([x.replace('/', os.sep) for x in result])


def test_action_option_template(clear_now_and_after):
    clear_now_and_after('template_output.txt')

    execute_workflow('''
        run:  template='cat {filename}'
            echo 'whatever' > template_output.txt
        ''')
    assert not os.path.isfile('template_output.txt')


def test_action_option_template_name(config_factory):
    cfg = config_factory('''\
        action_templates:
            cat: |
                cat {filename}

        ''')
    execute_workflow(
        '''
        run:  template_name='cat'
            echo 'whatever' > template_output.txt
        ''',
        args=['-c', cfg])
    assert not os.path.isfile('template_output.txt')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import glob
import multiprocessing as mp
import os
import shutil
import socket
import stat
import subprocess
import sys
import pexpect
from collections.abc import Sequence

import pkg_resources

from .eval import Undetermined, cfg_interpolate
from .syntax import SOS_LOGLINE
from .targets import path, sos_targets
from .task_engines import BackgroundProcess_TaskEngine
from .workflow_engines import BackgroundProcess_WorkflowEngine
from .tasks import TaskFile
from .utils import (env, expand_size, expand_time, format_HHMMSS, short_repr)

#
# A 'queue' is defined by queue configurations in SoS configuration files.
# It encapsulate properties of a queue and tells sos how to interact with
# the queue. A queue can be a local host, a remote host without queue, or
# a remote host with a task queue, or a RQ or Celery server. There are
# two main categories of properties of a host.
#
# 1. host properties, namely how to copy files and how to execute comamnds
#   on the host. Note that even for queues that communicate with sos
#   through network, the workers might be on a different host with different
#   file systems.
#
#   Keys for host configuration include:
#   * path_map: path map between local and remote hosts
#   * shared: paths that are shared between local and remote hosts
#   * send_cmd (optional): alternative command to send files
#   * receive_cmd (optional): alternative command to receive files
#   * execute_cmd (optional): alternative command to execute commands
#
# 2. task properties, namely how to manage running jobs. These include
#   direct execution, PBS and various cluster systems, and various task
#   queues.
#
#   Keys for task configuration depend largely one queue type.
#
#   * task_engine: type of task engine
#   * max_jobs: maximum number of concurrent jobs on the host.
#
#
# Implementation wise, a queue instance is created for each queue.
#

from typing import Any, Dict, List, Optional, Union


class DaemonizedProcess(mp.Process):

    def __init__(self, cmd, *args, **kwargs):
        super(DaemonizedProcess, self).__init__(*args, **kwargs)
        self.cmd = cmd

    def run(self):
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as err:
            env.logger.error(f'_Fork #1 failed: {err}\n')
            sys.exit(1)

        os.setsid()
        os.umask(0)
        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as err:
            env.logger.error(f'_Fork #2 failed: {err}\n')
            sys.exit(1)
        # the following is also need to properly daemonize the process
        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        try:
            si = open(os.devnull, 'r')
            so = open(os.devnull, 'w')
            se = open(os.devnull, 'w')
            os.dup2(si.fileno(), sys.stdin.fileno())
            os.dup2(so.fileno(), sys.stdout.fileno())
            os.dup2(se.fileno(), sys.stderr.fileno())
        except Exception:
            # #493
            pass

        # fork a new process
        subprocess.Popen(self.cmd, shell=True, close_fds=True)
        return


def _show_err_and_out(task_id, res) -> None:
    if 'stdout' in res:
        sys.stderr.write(f'\n{task_id}.out:\n')
        ends_with_newline = False
        for line in res['stdout'].splitlines():
            if not SOS_LOGLINE.match(line):
                sys.stderr.write(line)
                ends_with_newline = line.endswith('\n')
        if not ends_with_newline:
            sys.stderr.write('\n')
    if 'stderr' in res:
        sys.stderr.write(f'\n{task_id}.err:\n')
        ends_with_newline = False
        for line in res['stderr'].splitlines():
            if not SOS_LOGLINE.match(line):
                sys.stderr.write(line)
                ends_with_newline = line.endswith('\n')
        if not ends_with_newline:
            sys.stderr.write('\n')


class LocalHost(object):
    '''For local host, no path map, send and receive ...'''

    def __init__(self,
                 config: Dict[str, Union[str, int, List[str]]],
                 test_connection: bool = True) -> None:
        super(LocalHost, self).__init__()

        # even if the config has an alias, we use localhost to make it clear that the host is localhost
        self.alias = config.get('alias', 'localhost')
        self.address = 'localhost'
        # we checkk local jobs more aggressively
        self.config = {'alias': self.alias, 'status_check_interval': 2}
        self.config.update(config)

    def test_connection(self):
        return 'OK'

    def send_to_host(self, items):
        return {x: x for x in items}

    def receive_from_host(self, items):
        return {x: x for x in items}

    def prepare_task(self, task_id):
        task_file = os.path.join(
            os.path.expanduser('~'), '.sos', 'tasks', task_id + '.task')
        # add server restriction on task file
        if not os.path.isfile(task_file):
            raise ValueError(f'Missing task definition {task_file}')

        tf = TaskFile(task_id)
        params, old_runtime = tf.get_params_and_runtime()
        # clear possible previous result
        task_vars = params.sos_dict
        runtime = {
            '_runtime': {
                x: task_vars['_runtime'][x]
                for x in ('verbosity', 'sig_mode', 'run_mode', 'walltime',
                          'cores', 'mem')
                if x in task_vars['_runtime']
            }
        }
        runtime['_runtime']['workdir'] = task_vars['_runtime'][
            'workdir'] if 'workdir' in task_vars['_runtime'] else os.getcwd()

        if 'max_mem' in self.config or 'max_cores' in self.config or 'max_walltime' in self.config:
            for key in ('max_mem', 'max_cores', 'max_walltime'):
                if key in self.config:
                    runtime['_runtime'][key] = format_HHMMSS(
                        self.config[key]
                    ) if key == 'max_walltime' else self.config[key]

            if self.config.get('max_mem', None) is not None and task_vars['_runtime'].get('mem', None) is not None \
                    and self.config['max_mem'] < task_vars['_runtime']['mem']:
                env.logger.error(
                    f'Task {task_id} requested more mem ({task_vars["_runtime"]["mem"]}) than allowed max_mem ({self.config["max_mem"]})'
                )
                return False
            if self.config.get('max_cores', None) is not None and task_vars['_runtime'].get('cores', None) is not None \
                    and self.config['max_cores'] < task_vars['_runtime']['cores']:
                env.logger.error(
                    f'Task {task_id} requested more cores ({task_vars["_runtime"]["cores"]}) than allowed max_cores ({self.config["max_cores"]})'
                )
                return False
            if self.config.get('max_walltime', None) is not None and task_vars['_runtime'].get('walltime', None) is not None \
                    and expand_time(self.config['max_walltime']) < expand_time(task_vars['_runtime']['walltime']):
                env.logger.error(
                    f'Task {task_id} requested more walltime ({task_vars["_runtime"]["walltime"]}) than allowed max_walltime ({self.config["max_walltime"]})'
                )
                return False

        # if the task has been running remotely, we need to reset runtime for local execution
        if len(runtime) > 1 or runtime['_runtime'] or runtime != old_runtime:
            tf.runtime = runtime
        tf.status = 'pending'
        self.send_job_file(task_file)
        return True

    def send_job_file(self, job_file, dir='tasks'):
        # on the same file system, no action is needed.
        dest_job_file = os.path.join(
            os.path.expanduser('~'), '.sos', dir, os.path.basename(job_file))
        if job_file != dest_job_file:
            shutil.copyfile(job_file, dest_job_file)

    def check_output(self, cmd, under_workdir=False, **kwargs):
        # get the output of command
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        try:
            cmd = cfg_interpolate(cmd)
            return subprocess.check_output(
                cmd, shell=isinstance(cmd, str), **kwargs).decode()
        except Exception as e:
            env.logger.warning(f'Check output of {cmd} failed: {e}')
            raise

    def check_call(self, cmd, under_workdir=False, **kwargs):
        # get the output of command
        try:
            return subprocess.check_call(
                cmd, shell=isinstance(cmd, str), **kwargs)
        except Exception as e:
            env.logger.warning(f'Check output of {cmd} failed: {e}')
            raise

    def run_command(self, cmd, wait_for_task, realtime=False, **kwargs):
        # run command but does not wait for result.
        if realtime:
            from .utils import pexpect_run
            return pexpect_run(cmd)
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        if wait_for_task or sys.platform == 'win32':
            p = subprocess.Popen(cmd, shell=True, **kwargs)
            p.wait()
        else:
            p = DaemonizedProcess(cmd, **kwargs)
            p.start()
            p.join()

    def receive_result(self, task_id: str) -> Dict[str, Any]:
        tf = TaskFile(task_id)

        res = tf.result
        if not res or 'ret_code' not in res:
            return {
                'ret_code':
                    1,
                'task':
                    task_id,
                'exception':
                    ValueError(f'No result is received for task {task_id}')
            }

        try:
            if res['ret_code'] != 0 or env.verbosity >= 3:
                _show_err_and_out(task_id, res)
        except Exception as e:
            # if ret_code does not exist...
            return {'ret_code': 1, 'output': {}, 'exception': e}
        return res


class RemoteHost(object):
    '''A remote host class that manages how to communicate with remote host'''

    def __init__(self,
                 config: Dict[str, Union[str, int, List[str]]],
                 test_connection: bool = True) -> None:
        self.config = config
        self.cm_opts = self._get_control_master_options()
        self.alias = self.config['alias']
        #
        self.address = self.config['address']
        self.port = self.config.get('port', 22)
        self.shared_dirs = self._get_shared_dirs()
        self.path_map = self._get_path_map()
        # we already test connect of remote hosts
        if test_connection:
            test_res = self.test_connection()
            if test_res != 'OK':
                raise RuntimeError(
                    f'Failed to connect to {self.alias}: {test_res}')

    def _get_shared_dirs(self) -> List[Any]:
        value = self.config.get('shared', [])
        if isinstance(value, str):
            return [value]
        if isinstance(value, Sequence):
            return value
        raise ValueError(
            'Option shared can only be a string or a list of strings')

    def _get_path_map(self) -> Dict[str, str]:
        res: Dict = {}
        # if user-specified path_map, it overrides CONFIG
        path_map = self.config.get('path_map', [])
        #
        if not path_map:
            return res
        if isinstance(path_map, str):
            path_map = [path_map]
        if isinstance(path_map, Sequence):
            for v in path_map:
                if ' -> ' not in v:
                    raise ValueError(
                        f'Path map should be separated as from -> to, {v} specified'
                    )
                elif v.count(' -> ') > 1:
                    raise ValueError(
                        f'Path map should be separated as from -> to, {v} specified'
                    )
                res[v.split(' -> ')[0]] = v.split(' -> ')[1]
        elif isinstance(path_map, dict):
            for k, v in path_map.items():
                res[k] = v
        else:
            raise ValueError(
                f'Unacceptable path_mapue for configuration path_map: {path_map}'
            )
        return res

    def _get_control_master_options(self):
        master_dir = os.path.join(
            os.path.expanduser('~'), '.ssh', 'controlmasters')
        if not os.path.isdir(master_dir):
            try:
                os.makedirs(master_dir, exist_ok=True)
            except Exception as e:
                env.logger.debug(
                    f'Failed to create ssh control master directory {master_dir}: {e}'
                )
                return ''
        return f'-o "ControlMaster=auto" -o "ControlPath={master_dir}/%r@%h:%p" -o "ControlPersist=10m"'

    def _get_send_cmd(self, rename=False):
        if rename:
            return 'ssh ' + self.cm_opts + ''' -q {host} -p {port} "mkdir -p {dest:dpq}" && ''' + \
                '''rsync -a --no-g -e 'ssh ''' + self.cm_opts + ''' -p {port}' {source:aep} "{host}:{dest:dep}" && ''' + \
                '''ssh ''' + self.cm_opts + ''' -q {host} -p {port} "mv {dest:dep}/{source:b} {dest:ep}" '''
        return 'ssh ' + self.cm_opts + ''' -q {host} -p {port} "mkdir -p {dest:dpq}" && rsync -a --no-g -e 'ssh -p {port}' {source:aep} "{host}:{dest:dep}"'''

    def _get_receive_cmd(self, rename=False):
        if rename:
            return '''rsync -a --no-g -e 'ssh ''' + self.cm_opts + ''' -p {port}' {host}:{source:e} "{dest:adep}" && ''' + \
                '''mv "{dest:adep}/{source:b}" "{dest:aep}"'''
        return '''rsync -a --no-g -e 'ssh ''' + self.cm_opts + ''' -p {port}' {host}:{source:e} "{dest:adep}"'''

    def _get_execute_cmd(self, under_workdir=True) -> str:
        return self.config.get(
            'execute_cmd', 'ssh ' + self.cm_opts +
            """ -q {host} -p {port} "bash --login -c '""" +
            (' [ -d {workdir} ] || mkdir -p {workdir}; cd {workdir} && '
             if under_workdir else ' ') + ''' {cmd}'" ''')

    def _get_query_cmd(self):
        return self.config.get(
            'query_cmd', '''ssh ''' + self.cm_opts +
            ''' -q {host} -p {port} "bash --login -c 'sos status {task} -v 0'" '''
        )

    def is_shared(self, path):
        fullpath = os.path.abspath(os.path.expanduser(path))
        for sdir in self.shared_dirs:
            if fullpath.startswith(sdir):
                # issue 710, if a directory is both under path_map and shared, then it is not considered to be shared.
                if not any(
                        fullpath.startswith(mdir)
                        for mdir in self.path_map.keys()):
                    return True
        return False

    def _map_path(self, source):
        result = {}
        cwd = os.getcwd()
        if isinstance(source, int) and source.startswith('#'):
            result[path(source, host='localhost')] = path(
                source, host=self.alias)
        elif isinstance(source, (str, path)):
            dest = os.path.abspath(os.path.expanduser(source))
            # we use samefile to avoid problems with case-insensitive file system #522
            # we also use the "cwd" name to avoid wrong case for cwd. For example,
            # if the cwd = '/Users/user/Project'
            # then, dest = '/USERS/USER/PROJECT/a.txt'
            # would be converted to '/Users/user/Project/a.txt' before path mapping
            if os.path.exists(dest[:len(cwd)]) and os.path.samefile(
                    dest[:len(cwd)], cwd):
                dest = cwd + dest[len(cwd):]
            matched = [
                k for k in self.path_map.keys()
                if os.path.exists(dest[:len(k)]) and
                os.path.samefile(dest[:len(k)], k)
            ]
            if matched:
                # pick the longest key that matches
                k = max(matched, key=len)
                dest = self.path_map[k] + dest[len(k):]
            else:
                env.logger.debug(
                    f'Path {source} is not under any specified paths of localhost and is mapped to {dest} on remote host.'
                )
            result[source] = dest.replace('\\', '/')
        elif isinstance(source, (Sequence, sos_targets)):
            for src in source:
                result.update(self._map_path(src))
        else:
            env.logger.debug(f'Ignore unmappable source {source}')
            return {source: source}
        return result

    #
    # Interface functions
    #
    def _map_var(self, source):
        cwd = os.getcwd()
        if isinstance(source, path):
            source = str(source)
        if isinstance(source, str):
            dest = os.path.abspath(os.path.expanduser(source))
            # we use samefile to avoid problems with case-insensitive file system #522
            # we also use the "cwd" name to avoid wrong case for cwd. For example,
            # if the cwd = '/Users/user/Project'
            # then, dest = '/USERS/USER/PROJECT/a.txt'
            # would be converted to '/Users/user/Project/a.txt' before path mapping
            if os.path.exists(dest[:len(cwd)]) and os.path.samefile(
                    dest[:len(cwd)], cwd):
                dest = cwd + dest[len(cwd):]
            matched = [
                k for k in self.path_map.keys()
                if os.path.exists(dest[:len(k)]) and
                os.path.samefile(dest[:len(k)], k)
            ]
            if matched:
                # pick the longest key that matches
                k = max(matched, key=len)
                dest = self.path_map[k] + dest[len(k):]
            else:
                env.logger.debug(
                    f'Path {source} is not under any specified paths of localhost and is mapped to {dest} on remote host.'
                )
            return dest.replace('\\', '/')
        elif isinstance(source, (Sequence, set, sos_targets)):
            ret = [self._map_var(x) for x in source]
            return [x for x in ret if x is not None]
        else:
            env.logger.debug(f'Ignore unmappable source {source}')
            return source

    def test_connection(self):
        try:
            cmd = cfg_interpolate('ssh {host} -p {port} true', {
                'host': self.address,
                'port': self.port
            })
            p = pexpect.spawn(cmd)
            # could be prompted for Password or password, so use assword
            i = p.expect([
                "(?i)are you sure you want to continue connecting",
                "[pP]assword:", pexpect.EOF
            ],
                         timeout=5)
            if i == 0:
                p.sendline('yes')
                p.expect([
                    "(?i)are you sure you want to continue connecting",
                    "[pP]assword:", pexpect.EOF
                ],
                         timeout=5)
            elif i == 1:
                p.close(force=True)
                from .remote import stty_sane
                stty_sane()
                return f'ssh connection to {self.address} was prompted for password. Please set up public key authentication to the remote host before continue.'
            elif i == 2:
                p.close()
                if p.exitstatus == 0:
                    return 'OK'
                elif p.before:
                    return p.before.decode()
                else:
                    return f'Command "{cmd}" exits with code {p.exitstatus}'
        except pexpect.TIMEOUT:
            return f'ssh connection to {self.address} time out with prompt: {str(p.before)}'
        except Exception as e:
            return f'Failed to check remote connection {self.address}:{self.port}: {e}'
        return "OK"

    def _reverse_map_var(self, dest):
        if isinstance(dest, path):
            dest = str(dest)
        if isinstance(dest, str):
            matched = [
                l for l, r in self.path_map.items()
                if dest.startswith(r) and (len(r) == len(dest) or r.endswith(
                    '/') or r.endswith('\\') or dest[len(r)] in ('/', '\\'))
            ]
            if matched:
                # pick the longest key that matches
                k = max(matched, key=len)
                dest = k + dest[len(self.path_map[k]):]
            else:
                env.logger.debug(
                    f'Path {dest} is not under any specified paths of localhost and is mapped to {dest} on local host.'
                )
            return dest.replace('\\', '/')
        elif isinstance(dest, (Sequence, set, sos_targets)):
            ret = [self._reverse_map_var(x) for x in dest]
            return [x for x in ret if x is not None]
        else:
            env.logger.debug(f'Ignore unmappable source {dest}')
            return dest

    def _remote_abs(self, path):
        # return an absolute path relative to remote host
        path = str(path)
        if os.path.isabs(path):
            return path
        return os.path.join(self._map_var(os.getcwd()), path)

    def send_to_host(self, items):
        # we only copy files and directories, not other types of targets
        if isinstance(items, str):
            items = glob.glob(items)
        elif isinstance(items, path):
            items = [str(items)]
        elif isinstance(items, Sequence):
            ignored = [x for x in items if not isinstance(x, (str, path))]
            if ignored:
                env.logger.info(f'``Ignore`` {ignored}')
            items = sum([
                glob.glob(x) if isinstance(x, str) else [x]
                for x in items
                if x not in ignored
            ], [])
        else:
            env.logger.warning(
                f'Unrecognized items to be sent to host: {items}')
            return {}

        from .utils import find_symbolic_links
        new_items = []
        for item in items:
            links = find_symbolic_links(item)
            for link, realpath in links.items():
                env.logger.info(f'Adding {realpath} for symbolic link {link}')
            new_items.extend(links.values())
        items.extend(new_items)

        sending = self._map_path(items)

        sent = {}
        for source in sorted(sending.keys()):
            dest = self._remote_abs(sending[source])
            if self.is_shared(source):
                if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    env.log_to_file(
                        'TASK', f'Skip sending {source} on shared file system')
            else:
                if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    env.log_to_file(
                        'TASK', f'Sending ``{source}`` to {self.alias}:{dest}')
                cmd = cfg_interpolate(
                    self._get_send_cmd(
                        rename=os.path.basename(source) != os.path.basename(
                            dest)), {
                                'source': sos_targets(str(source).rstrip('/')),
                                'dest': sos_targets(dest),
                                'host': self.address,
                                'port': self.port
                            })
                if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    env.log_to_file('TASK', cmd)
                ret = subprocess.call(
                    cmd,
                    shell=True,
                    stderr=subprocess.DEVNULL,
                    stdout=subprocess.DEVNULL)
                if (ret != 0):
                    raise RuntimeError(
                        f'Failed to copy {source} to {self.alias} using command "{cmd}". The remote host might be unavailable.'
                    )
            sent[source] = dest
        return sent

    def receive_from_host(self, items):
        if isinstance(items, str):
            items = [items]
        elif isinstance(items, path):
            items = [str(items)]
        elif isinstance(items, Sequence):
            ignored = [x for x in items if not isinstance(x, (str, path))]
            if ignored:
                env.logger.info(f'``Ignore`` {ignored}')
            items = [x for x in items if isinstance(x, (str, path))]
        else:
            env.logger.warning(
                f'Unrecognized items to be retrieved from host: {items}')
            return {}

        # y could be path
        receiving = {
            self._remote_abs(y): str(x)
            for x, y in self._map_path(items).items()
        }
        #
        received = {}
        for source in sorted(receiving.keys()):
            dest = receiving[source]
            dest_dir = os.path.dirname(dest)
            if dest_dir and not os.path.isdir(dest_dir):
                try:
                    os.makedirs(dest_dir)
                except Exception as e:
                    env.logger.error(
                        f'Failed to create destination directory {dest_dir}: {e}'
                    )
            if self.is_shared(dest) and os.path.basename(
                    source) == os.path.basename(dest):
                env.logger.debug(
                    f'Skip retrieving ``{dest}`` from shared file system')
                received[dest] = source
            else:
                cmd = cfg_interpolate(
                    self._get_receive_cmd(
                        rename=os.path.basename(source) != os.path.basename(
                            dest)), {
                                'source': sos_targets(str(source).rstrip('/')),
                                'dest': sos_targets(dest),
                                'host': self.address,
                                'port': self.port
                            })
                if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    env.log_to_file('TASK', cmd)
                try:
                    ret = subprocess.call(
                        cmd,
                        shell=True,
                        stderr=subprocess.DEVNULL,
                        stdout=subprocess.DEVNULL)
                    if (ret != 0):
                        raise RuntimeError(f'command return {ret}')
                    received[dest] = source
                except Exception as e:
                    raise RuntimeError(
                        f'Failed to copy {source} from {self.alias} using command "{cmd}": {e}'
                    )
        return received

    #
    # Interface
    #
    def prepare_task(self, task_id):
        try:
            self._prepare_task(task_id)
            return True
        except Exception as e:
            env.logger.error(str(e))
            return False

    def _prepare_task(self, task_id):
        task_file = os.path.join(
            os.path.expanduser('~'), '.sos', 'tasks', task_id + '.task')
        if not os.path.isfile(task_file):
            raise ValueError(f'Missing task definition {task_file}')
        tf = TaskFile(task_id)
        params, old_runtime = tf.get_params_and_runtime()
        task_vars = params.sos_dict
        runtime = {
            '_runtime': {
                x: task_vars['_runtime'][x]
                for x in ('verbosity', 'sig_mode', 'run_mode', 'walltime',
                          'cores', 'mem')
                if x in task_vars['_runtime']
            },
            task_id: {}
        }

        if self.config.get('max_mem', None) is not None and task_vars['_runtime'].get('mem', None) is not None \
                and self.config['max_mem'] < task_vars['_runtime']['mem']:
            raise ValueError(
                f'Task {task_id} requested more mem ({task_vars["_runtime"]["mem"]}) than allowed max_mem ({self.config["max_mem"]})'
            )
        if self.config.get('max_cores', None) is not None and task_vars['_runtime'].get('cores', None) is not None \
                and self.config['max_cores'] < task_vars['_runtime']['cores']:
            raise ValueError(
                f"Task {task_id} requested more cores ({task_vars['_runtime']['cores']}) than allowed max_cores ({self.config['max_cores']})"
            )
        if self.config.get('max_walltime', None) is not None and task_vars['_runtime'].get('walltime', None) is not None \
                and expand_time(self.config['max_walltime']) < expand_time(task_vars['_runtime']['walltime']):
            raise ValueError(
                f'Task {task_id} requested more walltime ({task_vars["_runtime"]["walltime"]}) than allowed max_walltime ({self.config["max_walltime"]})'
            )

        if task_vars['_input'] and not isinstance(task_vars['_input'],
                                                  Undetermined):
            sent = self.send_to_host(task_vars['_input'])
            if sent:
                env.logger.info(
                    f'{task_id} ``sent`` {short_repr(sent.keys())} to {self.alias}'
                )
        if task_vars['_depends'] and not isinstance(task_vars['_depends'],
                                                    Undetermined):
            sent = self.send_to_host(task_vars['_depends'])
            if sent:
                env.logger.info(
                    f'{task_id} ``sent`` {short_repr(sent.keys())} to {self.alias}'
                )
        if 'to_host' in task_vars['_runtime']:
            if not isinstance(task_vars['_runtime']['to_host'], (str, Sequence)) or \
                (isinstance(task_vars['_runtime']['to_host'], Sequence) and \
                    not all(isinstance(x, str) for x in task_vars['_runtime']['to_host'])):
                raise ValueError(
                    f'Parameter to_host accepts a list of paths (strings). {task_vars["_runtime"]["to_host"]} provided'
                )

            sent = self.send_to_host(task_vars['_runtime']['to_host'])
            if sent:
                env.logger.info(
                    f'{task_id} ``sent`` {short_repr(sent.keys())} to {self.alias}'
                )

        # map variables
        if 'workdir' in task_vars['_runtime']:
            runtime['_runtime']['workdir'] = self._map_var(
                task_vars['_runtime']['workdir'])
        else:
            runtime['_runtime']['workdir'] = self._map_var(os.getcwd())

        mapped_vars = {
            '_input', '_output', '_depends', 'input', 'output', 'depends'
        }

        for var in mapped_vars:
            if var not in task_vars:
                # input, output, depends might not exist
                continue
            if not task_vars[var]:
                continue
            elif isinstance(task_vars[var], str):
                runtime[task_id][var] = self._map_var(task_vars[var])
                env.log_to_file(
                    'TASK',
                    f'On {self.alias}: ``{var}`` = {short_repr(task_vars[var])}'
                )
            elif isinstance(task_vars[var], (Sequence, set)):
                runtime[task_id][var] = type(task_vars[var])(
                    self._map_var(task_vars[var]))
                env.log_to_file(
                    'TASK',
                    f'On {self.alias}: ``{var}`` = {short_repr(task_vars[var])}'
                )
            else:
                env.logger.warning(
                    f'Failed to map {var} of type {task_vars[var].__class__.__name__}'
                )

        # master task??
        if hasattr(params, 'task_stack'):
            for tid, tdef in params.task_stack:
                runtime[tid] = {}
                for var in mapped_vars:
                    if var not in tdef.sos_dict:
                        # input, output, depends might not exist
                        continue
                    if not tdef.sos_dict[var]:
                        continue
                    elif isinstance(tdef.sos_dict[var], str):
                        runtime[tid][var] = self._map_var(tdef.sos_dict[var])
                    elif isinstance(tdef.sos_dict[var], (Sequence, set)):
                        runtime[tid][var] = type(tdef.sos_dict[var])(
                            self._map_var(tdef.sos_dict[var]))
                    else:
                        env.logger.warning(
                            f'Failed to map {var} of type {tdef.sos_dict[var].__class__.__name__}'
                        )

        # server restrictions #488
        for key in ('max_mem', 'max_cores', 'max_walltime'):
            if key in self.config:
                runtime['_runtime'][key] = format_HHMMSS(
                    self.config[key]
                ) if key == 'max_walltime' else self.config[key]

        # only update task file if there are runtime information
        if len(runtime) > 1 or runtime['_runtime'] or runtime != old_runtime:
            tf.runtime = runtime

        tf.status = 'pending'
        self.send_job_file(task_file)

    def send_job_file(self, job_file, dir='tasks'):
        send_cmd = cfg_interpolate(
            f'ssh {self.cm_opts}'
            f' -q {{address}} -p {{port}} "[ -d ~/.sos/{dir} ] || mkdir -p ~/.sos/{dir}" && '
            f' rsync --ignore-existing -a --no-g -e "ssh {self.cm_opts}'
            f' -q -p {{port}}" {{job_file:ap}} {{address}}:.sos/{dir}/', {
                'job_file': sos_targets(job_file),
                'address': self.address,
                'port': self.port
            })
        # use scp for this simple case
        try:
            subprocess.check_call(send_cmd, shell=True)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f'Failed to copy job {job_file} to {self.alias} using command {send_cmd}: {e}'
            )

    def check_output(self, cmd: object, under_workdir=False,
                     **kwargs) -> object:
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        try:
            cmd = cfg_interpolate(
                self._get_execute_cmd(under_workdir=under_workdir), {
                    'host': self.address,
                    'port': self.port,
                    'cmd': cmd,
                    'workdir': self._map_var(os.getcwd())
                })
        except Exception as e:
            raise ValueError(
                f'Failed to run command {cmd}: {e} ({env.sos_dict["CONFIG"]})')
        if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
            env.log_to_file('TASK', f'Executing command ``{cmd}``')
        try:
            return subprocess.check_output(cmd, shell=True, **kwargs).decode()
        except Exception as e:
            env.logger.debug(f'Check output of {cmd} failed: {e}')
            raise

    def check_call(self, cmd, under_workdir=False, **kwargs):
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        try:
            cmd = cfg_interpolate(
                self._get_execute_cmd(under_workdir=under_workdir), {
                    'host': self.address,
                    'port': self.port,
                    'cmd': cmd,
                    'workdir': self._map_var(os.getcwd())
                })
        except Exception as e:
            raise ValueError(f'Failed to run command {cmd}: {e}')
        if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
            env.log_to_file('TASK', f'Executing command ``{cmd}``')
        try:
            return subprocess.check_call(cmd, shell=True, **kwargs)
        except Exception as e:
            env.logger.debug(f'Check output of {cmd} failed: {e}')
            raise

    def run_command(self, cmd, wait_for_task, realtime=False, **kwargs):
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        try:
            cmd = cfg_interpolate(
                self._get_execute_cmd(under_workdir=False), {
                    'host': self.address,
                    'port': self.port,
                    'cmd': cmd,
                    'workdir': self._map_var(os.getcwd())
                })
        except Exception as e:
            raise ValueError(f'Failed to run command {cmd}: {e}')
        if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
            env.log_to_file('TASK', f'Executing command ``{cmd}``')
        if realtime:
            from .utils import pexpect_run
            return pexpect_run(cmd)
        elif wait_for_task or sys.platform == 'win32':
            # keep proc persistent to avoid a subprocess is still running warning.
            p = subprocess.Popen(cmd, shell=True, **kwargs)
            p.wait()
        else:
            p = DaemonizedProcess(cmd, **kwargs)
            p.start()
            p.join()

    def receive_result(self, task_id: str) -> Dict[str, int]:
        # for filetype in ('res', 'status', 'out', 'err'):
        sys_task_dir = os.path.join(os.path.expanduser('~'), '.sos', 'tasks')
        # use -p to preserve modification times so that we can keep the job status locally.
        receive_cmd = cfg_interpolate(
            "scp -P {port} -p -q {address}:.sos/tasks/{task}.* {sys_task_dir}",
            {
                'port': self.port,
                'address': self.address,
                'task': task_id,
                'sys_task_dir': sys_task_dir
            })
        # it is possible that local files are readonly (e.g. a pluse file) so we first need to
        # make sure the files are readable and remove them. Also, we do not want any file that is
        # obsolete to appear as new after copying
        for lfile in glob.glob(
                os.path.join(
                    os.path.expanduser('~'), '.sos', 'tasks', task_id + '.*')):
            if not os.access(lfile, os.W_OK):
                os.chmod(lfile, stat.S_IREAD | stat.S_IWRITE)
            #os.remove(lfile)
        if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
            env.log_to_file('TASK', receive_cmd)
        ret = subprocess.call(
            receive_cmd,
            shell=True,
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL)
        if (ret != 0):
            # this time try to get error message
            ret = subprocess.call(receive_cmd, shell=True)
            if (ret != 0):
                raise RuntimeError(
                    'Failed to retrieve result of job {} from {} with cmd\n{}'
                    .format(task_id, self.alias, receive_cmd))

        tf = TaskFile(task_id)
        params = tf.params
        res = tf.result

        if not res:
            env.logger.debug(
                f'Result for {task_id} is not received (no result)')
            return {
                'ret_code': 1,
                'task': task_id,
                'exception': RuntimeError(f'Task {task_id} failed or aborted'),
                'output': sos_targets()
            }

        if ('ret_code' in res and res['ret_code'] != 0) or ('succ' in res and
                                                            res['succ'] != 0):
            _show_err_and_out(task_id, res)
            env.logger.info(f'Ignore remote results for failed job {task_id}')
            return res

        if env.verbosity >= 3:
            _show_err_and_out(task_id, res)
        # do we need to copy files? We need to consult original task file
        # not the converted one
        job_dict = params.sos_dict
        if '_output' in job_dict and job_dict['_output'] and not isinstance(
                job_dict['_output'],
                Undetermined) and env.config['run_mode'] != 'dryrun':
            received = self.receive_from_host(
                [x for x in job_dict['_output'] if isinstance(x, (str, path))])
            if received:
                env.logger.info(
                    f'{task_id} ``received`` {short_repr(received.keys())} from {self.alias}'
                )
        if 'from_host' in job_dict[
                '_runtime'] and env.config['run_mode'] != 'dryrun':
            if isinstance(job_dict['_runtime']['from_host'], (Sequence, str)):
                received = self.receive_from_host(
                    job_dict['_runtime']['from_host'])
                if received:
                    env.logger.info(
                        f'{task_id} ``received`` {short_repr(received.keys())} from {self.alias}'
                    )
            else:
                env.logger.warning(
                    f"Expecting a string or list of string from from_host: {job_dict['_runtime']['from_host']} received"
                )
        # we need to translate result from remote path to local
        if 'output' in res:
            if '_output' not in job_dict:
                # this should exist, but let us check it for robustness
                env.logger.warning('Missing _output in task dict')
                res['output'] = sos_targets()
            elif job_dict['_output'].undetermined():
                res['output'] = sos_targets(
                    self._reverse_map_var(res['output']))
            else:
                res['output'] = job_dict['_output']
        if 'subtasks' in res:
            for tid, subparams in params.task_stack:
                if tid in res['subtasks'] and 'output' in res['subtasks'][tid]:
                    if '_output' not in subparams.sos_dict:
                        # this should not happen
                        env.logger.warning('Missing _output in subparams')
                        res['subtasks'][tid]['output'] = sos_targets()
                    elif subparams.sos_dict['_output'].undetermined():
                        res['subtasks'][tid]['output'] = sos_targets(
                            self._reverse_map_var(
                                res['subtasks'][tid]['output']))
                    else:
                        res['subtasks'][tid]['output'] = subparams.sos_dict[
                            '_output']
        return res


#
# host instances are shared by all tasks so there should be only one
# instance for each host.
#


class Host:
    host_instances: Dict = {}

    def __init__(self,
                 alias: Optional[str] = '',
                 start_engine: bool = True,
                 test_connection: bool = True) -> None:
        # a host started from Jupyter notebook might not have proper stdout
        # (a StringIO) and cannot be used for DaemonizedFork).
        self._get_config(alias)
        self._get_host_agent(start_engine, test_connection)

    # for test purpose
    @classmethod
    def reset(cls) -> None:
        for host in cls.host_instances.values():
            # perhaps the engine failed to start
            if hasattr(host, '_task_engine'):
                del host._task_engine
        cls.host_instances = {}

    def _get_local_host(self) -> str:
        if 'CONFIG' not in env.sos_dict or 'hosts' not in env.sos_dict['CONFIG']:
            from .utils import load_config_files
            load_config_files()
        # look for an entry with gethost
        if 'hosts' not in env.sos_dict['CONFIG']:
            env.sos_dict['CONFIG']['hosts'] = {
                'localhost': {
                    'address': 'localhost',
                    'alias': 'localhost'
                }
            }
            return 'localhost'
        # try host name
        hostname = socket.gethostname().lower()
        for host, host_info in env.sos_dict['CONFIG']['hosts'].items():
            # find by key hostname
            if 'hostname' in host_info and host_info['hostname'].lower(
            ) == hostname:
                return host
        for host, host_info in env.sos_dict['CONFIG']['hosts'].items():
            # find by key hostname
            if 'hostname' in host_info and (
                    host_info['hostname'].lower().split('.')[0] == hostname or
                    host_info['hostname'].lower() == hostname.split('.')[0]):
                return host
        for host, host_info in env.sos_dict['CONFIG']['hosts'].items():
            # find by alias
            if host.lower() == hostname:
                return host
        for host, host_info in env.sos_dict['CONFIG']['hosts'].items():
            # find by address
            if 'address' in host_info and (
                    host_info['address'].split('@')[-1].lower() == hostname or
                    host_info['address'].split(
                        '.', 1)[0].split('@')[-1].lower() == hostname):
                return host
        # try IP Address
        hostname = socket.gethostname()
        ips = socket.gethostbyname_ex(hostname)[2]
        ips = [ip for ip in ips if not ip.startswith("127.")]
        # if the IP matches any address?
        for host, host_info in env.sos_dict['CONFIG']['hosts'].items():
            # find by key hostname
            if 'address' in host_info and any(
                    ip == host_info['address'].split('@')[-1] for ip in ips):
                return host
        # now check if a key localhost is defined
        if 'localhost' in env.sos_dict['CONFIG']:
            if env.sos_dict['CONFIG']['localhost'] not in env.sos_dict[
                    'CONFIG']['hosts']:
                raise ValueError(
                    f"Undefined localhost {env.sos_dict['CONFIG']['localhost']}"
                )
            return env.sos_dict['CONFIG']['localhost']
        else:
            env.sos_dict['CONFIG']['localhost'] = 'localhost'
            return 'localhost'

    def _get_remote_host(self, alias: Optional[str]) -> str:
        # get a remote host specified by Alias
        if 'CONFIG' not in env.sos_dict or 'hosts' not in env.sos_dict['CONFIG']:
            from .utils import load_config_files
            load_config_files()
        if not alias or alias == 'localhost':
            return self._get_local_host()
        if not isinstance(alias, str):
            raise ValueError(f'A string is expected for host {alias}')
        if 'hosts' not in env.sos_dict['CONFIG']:
            env.sos_dict['CONFIG']['hosts'] = {}
        if alias in env.sos_dict['CONFIG']['hosts']:
            return alias
        # assuming the host name is a name or IP address
        env.logger.debug(
            f'Assuming {alias} to be a hostname or IP address not defined in hosts file'
        )
        env.sos_dict['CONFIG']['hosts'][alias] = {
            'address': alias,
            'alias': alias,
        }
        return alias

    def _get_config(self, alias: Optional[str]) -> None:
        LOCAL = self._get_local_host()
        REMOTE = self._get_remote_host(alias)
        self.alias = REMOTE

        # now we need to find definition for local and remote host
        if LOCAL == 'localhost' and REMOTE == 'localhost':
            self.config = {
                'address': 'localhost',
                'alias': 'localhost',
            }
        elif 'hosts' in env.sos_dict['CONFIG']:
            if LOCAL not in env.sos_dict['CONFIG']['hosts']:
                raise ValueError(f'Undefined local host {LOCAL}')
            if REMOTE not in env.sos_dict['CONFIG']['hosts']:
                raise ValueError(f'Undefined remote host {REMOTE}')

            # now we have definition for local and remote hosts
            cfg = env.sos_dict['CONFIG']['hosts']
            self.config = {
                x: y
                for x, y in cfg[self.alias].items()
                if x not in ('paths', 'shared')
            }

            # if local and remote hosts are the same
            if LOCAL == REMOTE or 'address' not in env.sos_dict['CONFIG'][
                    'hosts'][REMOTE] or (
                        'address' in env.sos_dict['CONFIG']['hosts'][REMOTE] and
                        env.sos_dict['CONFIG']['hosts'][REMOTE]['address'] ==
                        'localhost'):
                # there would be no path map
                self.config['path_map'] = []
                self.config['shared'] = ['/']
                # override address setting to use localhost
                self.config['address'] = 'localhost'
            else:
                self.config['path_map'] = []

                def normalize_value(x):
                    x = cfg_interpolate(x)
                    return x if x.endswith(os.sep) else (x + os.sep)

                if 'shared' in cfg[LOCAL] and 'shared' in cfg[REMOTE]:
                    common = set(cfg[LOCAL]['shared'].keys()) & set(
                        cfg[REMOTE]['shared'].keys())
                    if common:
                        self.config['shared'] = [
                            normalize_value(cfg[LOCAL]['shared'][x])
                            for x in common
                        ]
                        self.config['path_map'] = [
                            f'{normalize_value(cfg[LOCAL]["shared"][x])} -> {normalize_value(cfg[REMOTE]["shared"][x])}'
                            for x in common
                        ]
                # if paths are defined for both local and remote host, define path_map
                if ('paths' in cfg[LOCAL] and
                        cfg[LOCAL]['paths']) and ('paths' in cfg[REMOTE] and
                                                  cfg[REMOTE]['paths']):
                    if any(k not in cfg[REMOTE]['paths']
                           for k in cfg[LOCAL]['paths'].keys()):
                        env.logger.debug(
                            f'One or more local paths {", ".join(cfg[LOCAL]["paths"].keys())} cannot be mapped to remote host {REMOTE} with paths {",".join(cfg[REMOTE]["paths"].keys())}'
                        )
                    #
                    self.config['path_map'].extend([
                        f'{normalize_value(cfg[LOCAL]["paths"][x])} -> {normalize_value(cfg[REMOTE]["paths"][x])}'
                        for x in cfg[LOCAL]['paths'].keys()
                        if x in cfg[REMOTE]['paths']
                    ])
        elif LOCAL == REMOTE:
            # now we have checked local and remote are not defined, but they are the same, so
            # it is safe to assume that they are both local hosts
            self.config = {
                'address': 'localhost',
                'alias': LOCAL,
            }
        else:
            raise ValueError(
                f'Undefined local and remote hosts {LOCAL} and {REMOTE}.')
        #
        self.config['alias'] = self.alias
        self.description = self.config.get('description', '')

        # standardize parameters max_walltime, max_cores, and max_mem for the host
        if 'max_walltime' in self.config:
            self.config['max_walltime'] = format_HHMMSS(
                self.config['max_walltime'])
        if 'max_cores' in self.config:
            if not isinstance(self.config['max_cores'], int):
                raise ValueError('An integer is expected for max_cores')
        if 'max_mem' in self.config:
            self.config['max_mem'] = expand_size(self.config['max_mem'])

    def _get_host_agent(self, start_engine: bool,
                        test_connection: bool) -> None:
        if 'queue_type' not in self.config:
            self._engine_type = 'process'
        else:
            self._engine_type = self.config['queue_type'].strip()
        if self.alias not in self.host_instances:
            if self.config['address'] == 'localhost':
                self.host_instances[self.alias] = LocalHost(
                    self.config, test_connection=test_connection)
            else:
                self.host_instances[self.alias] = RemoteHost(
                    self.config, test_connection=test_connection)

            if self._engine_type == 'process':
                task_engine = BackgroundProcess_TaskEngine(
                    self.host_instances[self.alias])
                workflow_engine = BackgroundProcess_WorkflowEngine(
                    self.host_instances[self.alias])
            else:
                task_engine = None
                workflow_engine = None

                for entrypoint in pkg_resources.iter_entry_points(
                        group='sos_taskengines'):
                    try:
                        if entrypoint.name == self._engine_type:
                            task_engine = entrypoint.load()(
                                self.host_instances[self.alias])
                            break
                    except Exception as e:
                        env.logger.debug(
                            f'Failed to load task engine {self._engine_type}: {e}'
                        )

                for entrypoint in pkg_resources.iter_entry_points(
                        group='sos_workflowengines'):
                    try:
                        if entrypoint.name == self._engine_type:
                            workflow_engine = entrypoint.load()(
                                self.host_instances[self.alias])
                            break
                    except Exception as e:
                        env.logger.debug(
                            f'Failed to load workflow engine {self._engine_type}: {e}'
                        )

            self.host_instances[self.alias]._task_engine = task_engine
            self.host_instances[self.alias]._workflow_engine = workflow_engine

            # the task engine is a thread and will run continously
            if start_engine and task_engine is not None:
                task_engine.start()

        self._host_agent = self.host_instances[self.alias]
        # for convenience
        if hasattr(self._host_agent, '_task_engine'):
            self._task_engine = self._host_agent._task_engine
        if hasattr(self._host_agent, '_workflow_engine'):
            self._workflow_engine = self._host_agent._workflow_engine
        # it is possible that Host() is initialized before with start_engine=False
        # and called again to start engine
        if start_engine and self._task_engine is not None and not self._task_engine.is_alive(
        ):
            self._task_engine.start()

    # public interface
    #
    # based on Host definition
    #
    def send_to_host(self, items):
        return self._host_agent.send_to_host(items)

    def receive_from_host(self, items):
        return self._host_agent.receive_from_host(items)

    def map_var(self, rvars):
        return self._host_agent._map_var(rvars)

    def submit_task(self, task_id: str) -> str:
        if not self._task_engine:
            raise RuntimeError(
                f'No task engine or invalid engine definition defined for host {self.alias}'
            )
        return self._task_engine.submit_task(task_id)

    def check_status(self, tasks: List[str]) -> List[str]:
        # find the task engine
        return [self._task_engine.check_task_status(task) for task in tasks]

    def retrieve_results(self, tasks: List[str]):
        return {task: self._host_agent.receive_result(task) for task in tasks}

    def execute_workflow(self, script, cmd, **template_args):
        if not self._workflow_engine:
            raise RuntimeError(
                f'No workflow engine or invalid engine definition defined for host {self.alias}'
            )
        return self._workflow_engine.execute_workflow(script, cmd,
                                                      **template_args)



import sys
import subprocess
import psutil


def run_process(args):
    proc_before = subprocess.check_output(
        'ps aux | grep -v root', shell=True).decode().splitlines()

    # run the main process
    ret = subprocess.Popen(args)
    ret.wait()

    proc_after = subprocess.check_output(
        'ps aux | grep -v root', shell=True).decode().splitlines()
    pid_before = [
        int(x.split()[1]) for x in proc_before if not 'PID' in x and 'TIME' in x
    ]
    pid_after = [
        int(x.split()[1]) for x in proc_after if not 'PID' in x and 'TIME' in x
    ]
    possible_zombies = [name for pid,name in zip(pid_after, proc_after) if pid not in pid_before and \
        psutil.Process(pid).ppid() not in pid_before]
    if possible_zombies:
        sys.exit('\nNew possible zoombie processes\n{zombies}')
    else:
        print('\nNo possible zombie process is detected')
        sys.exit(ret.returncode)


if __name__ == '__main__':
    if '-h' in sys.argv:
        print('Usage: python check_zombie.py regular command line')
        print(
            '''This command executes the command and lists new processes after the completion of '''
        )
        print(
            '''the command. Processes that are child processes of processes before execution are '''
        )
        print(
            '''excluded. The rest of the processes could be zombie process left by the command, or '''
        )
        print('''new processes created during the execution of the command.''')
    run_process(sys.argv[1:])



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

from sos.actions import SoS_Action, SoS_ExecuteScript


@SoS_Action(acceptable_args=['script', 'args'])
def bash(script, args='', **kwargs):
    '''Execute specified script using bash. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script.'''
    return SoS_ExecuteScript(script, ['/bin/bash', 'bash'], '.sh',
                             args).run(**kwargs)


@SoS_Action(acceptable_args=['script', 'args'])
def csh(script, args='', **kwargs):
    '''Execute specified script using csh. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script.'''
    return SoS_ExecuteScript(script, ['/bin/csh', 'csh'], '.csh',
                             args).run(**kwargs)


@SoS_Action(acceptable_args=['script', 'args'])
def tcsh(script, args='', **kwargs):
    '''Execute specified script using tcsh. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script.'''
    return SoS_ExecuteScript(script, ['/bin/tcsh', 'tcsh'], '.sh',
                             args).run(**kwargs)


@SoS_Action(acceptable_args=['script', 'args'])
def zsh(script, args='', **kwargs):
    '''Execute specified script using zsh. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script.'''
    return SoS_ExecuteScript(script, ['/bin/zsh', 'zsh'], '.zsh',
                             args).run(**kwargs)


@SoS_Action(acceptable_args=['script', 'args'])
def sh(script, args='', **kwargs):
    '''Execute specified script using sh. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script.'''
    return SoS_ExecuteScript(script, ['/bin/sh', 'sh'], '.sh',
                             args).run(**kwargs)



import pytest
import uuid
import os
import shutil
import tempfile
import textwrap
import pathlib


@pytest.fixture
def config_factory():
    filename = tempfile.NamedTemporaryFile(suffix='.yml', delete=False).name

    def get_config(text):
        with open(filename, 'w') as conf:
            conf.write(textwrap.dedent(text))
        return filename

    yield get_config
    os.remove(filename)


@pytest.fixture
def script_factory():
    filename = tempfile.NamedTemporaryFile(suffix='.sos', delete=False).name

    def get_script(text):
        with open(filename, 'w') as conf:
            conf.write(textwrap.dedent(text))
        return filename

    yield get_script
    os.remove(filename)


@pytest.fixture
def temp_factory():

    temp_fds = []
    temp_dirs = []

    def get_tempfiles(*args, **kwargs):
        for names in args:
            if isinstance(names, str):
                names = [names]
            for name in names:
                pathlib.Path(name).touch()
                temp_fds.append(name)
        if 'dir' in kwargs:
            if isinstance(kwargs['dir'], str):
                dirs = [kwargs['dir']]
            else:
                dirs = kwargs['dir']
            for dir in dirs:
                if os.path.isdir(dir):
                    shutil.rmtree(dir)
                os.makedirs(dir, exist_ok=True)
                temp_dirs.append(dir)
        return temp_fds, temp_dirs

    yield get_tempfiles

    for temp_file in temp_fds:
        try:
            if os.path.isfile(temp_file):
                os.remove(temp_file)
        except Exception:
            pass

    for temp_dir in temp_dirs:
        try:
            if os.path.isdir(temp_dir):
                shutil.rmtree(temp_dir)
        finally:
            pass


@pytest.fixture
def clear_now_and_after():
    # This fixture takes one or more file names or directory names and will
    # remove them if they exist, and also after the tests are completed.
    temp_fds = []

    def clear_files_and_dirs():
        for temp_fd in temp_fds:
            try:
                if os.path.isfile(temp_fd):
                    os.remove(temp_fd)
                elif os.path.isdir(temp_fd):
                    shutil.rmtree(temp_fd)
            except Exception:
                pass

    def get_names(*args):
        for names in args:
            if isinstance(names, str):
                temp_fds.append(names)
            else:
                temp_fds.extend(names)

        clear_files_and_dirs()
        return temp_fds

    yield get_names

    clear_files_and_dirs()
